@@ -1,3 +1,4 @@
+from __future__ import print_function
 from __future__ import with_statement
 
 from os.path import exists
@@ -12,10 +13,10 @@ from txtorcon import ITorControlProtocol
 from txtorcon.torcontrolprotocol import parse_keywords, DEFAULT_VALUE
 from txtorcon.util import hmac_sha256
 
-import types
 import functools
 import tempfile
 import base64
+from binascii import b2a_hex, a2b_hex
 
 
 class CallbackChecker:
@@ -26,7 +27,7 @@ class CallbackChecker:
     def __call__(self, *args, **kwargs):
         v = args[0]
         if v != self.expected_value:
-            print "WRONG"
+            print("WRONG")
             raise RuntimeError(
                 'Expected "%s" but got "%s"' % (self.expected_value, v)
             )
@@ -73,80 +74,99 @@ class AuthenticationTests(unittest.TestCase):
         self.transport = proto_helpers.StringTransport()
 
     def send(self, line):
-        self.protocol.dataReceived(line.strip() + "\r\n")
+        assert type(line) == bytes
+        self.protocol.dataReceived(line.strip() + b"\r\n")
 
     def test_authenticate_cookie(self):
         self.protocol.makeConnection(self.transport)
-        self.assertEqual(self.transport.value(), 'PROTOCOLINFO 1\r\n')
+        self.assertEqual(self.transport.value(), b'PROTOCOLINFO 1\r\n')
         self.transport.clear()
-        cookie_data = 'cookiedata!cookiedata!cookiedata'
-        with open('authcookie', 'w') as f:
+        cookie_data = b'cookiedata!cookiedata!cookiedata'
+        with open('authcookie', 'wb') as f:
             f.write(cookie_data)
-        self.send('250-PROTOCOLINFO 1')
-        self.send('250-AUTH METHODS=COOKIE,HASHEDPASSWORD COOKIEFILE="authcookie"')
-        self.send('250-VERSION Tor="0.2.2.34"')
-        self.send('250 OK')
+        self.send(b'250-PROTOCOLINFO 1')
+        self.send(b'250-AUTH METHODS=COOKIE,HASHEDPASSWORD COOKIEFILE="authcookie"')
+        self.send(b'250-VERSION Tor="0.2.2.34"')
+        self.send(b'250 OK')
 
         self.assertEqual(
             self.transport.value(),
-            'AUTHENTICATE %s\r\n' % cookie_data.encode("hex")
+            b'AUTHENTICATE ' + b2a_hex(cookie_data) + b'\r\n',
         )
 
     def test_authenticate_password(self):
         self.protocol.password_function = lambda: 'foo'
         self.protocol.makeConnection(self.transport)
-        self.assertEqual(self.transport.value(), 'PROTOCOLINFO 1\r\n')
+        self.assertEqual(self.transport.value(), b'PROTOCOLINFO 1\r\n')
         self.transport.clear()
-        self.send('250-PROTOCOLINFO 1')
-        self.send('250-AUTH METHODS=HASHEDPASSWORD')
-        self.send('250-VERSION Tor="0.2.2.34"')
-        self.send('250 OK')
+        self.send(b'250-PROTOCOLINFO 1')
+        self.send(b'250-AUTH METHODS=HASHEDPASSWORD')
+        self.send(b'250-VERSION Tor="0.2.2.34"')
+        self.send(b'250 OK')
 
-        self.assertEqual(self.transport.value(), 'AUTHENTICATE %s\r\n' % "foo".encode("hex"))
+        self.assertEqual(
+            self.transport.value(),
+            b'AUTHENTICATE ' + b2a_hex(b'foo') + b'\r\n'
+        )
+
+    def test_authenticate_password_not_bytes(self):
+        self.protocol.password_function = lambda: u'foo'
+        self.protocol.makeConnection(self.transport)
+        self.assertEqual(self.transport.value(), b'PROTOCOLINFO 1\r\n')
+        self.transport.clear()
+        self.send(b'250-PROTOCOLINFO 1')
+        self.send(b'250-AUTH METHODS=HASHEDPASSWORD')
+        self.send(b'250-VERSION Tor="0.2.2.34"')
+        self.send(b'250 OK')
+
+        self.assertEqual(
+            self.transport.value(),
+            b'AUTHENTICATE ' + b2a_hex(b'foo') + b'\r\n'
+        )
 
     def test_authenticate_null(self):
         self.protocol.makeConnection(self.transport)
-        self.assertEqual(self.transport.value(), 'PROTOCOLINFO 1\r\n')
+        self.assertEqual(self.transport.value(), b'PROTOCOLINFO 1\r\n')
         self.transport.clear()
-        self.send('250-PROTOCOLINFO 1')
-        self.send('250-AUTH METHODS=NULL')
-        self.send('250-VERSION Tor="0.2.2.34"')
-        self.send('250 OK')
+        self.send(b'250-PROTOCOLINFO 1')
+        self.send(b'250-AUTH METHODS=NULL')
+        self.send(b'250-VERSION Tor="0.2.2.34"')
+        self.send(b'250 OK')
 
-        self.assertEqual(self.transport.value(), 'AUTHENTICATE\r\n')
+        self.assertEqual(self.transport.value(), b'AUTHENTICATE\r\n')
 
     def test_authenticate_password_deferred(self):
         d = defer.Deferred()
         self.protocol.password_function = lambda: d
         self.protocol.makeConnection(self.transport)
-        self.assertEqual(self.transport.value(), 'PROTOCOLINFO 1\r\n')
+        self.assertEqual(self.transport.value(), b'PROTOCOLINFO 1\r\n')
         self.transport.clear()
-        self.send('250-PROTOCOLINFO 1')
-        self.send('250-AUTH METHODS=HASHEDPASSWORD')
-        self.send('250-VERSION Tor="0.2.2.34"')
-        self.send('250 OK')
+        self.send(b'250-PROTOCOLINFO 1')
+        self.send(b'250-AUTH METHODS=HASHEDPASSWORD')
+        self.send(b'250-VERSION Tor="0.2.2.34"')
+        self.send(b'250 OK')
 
         # make sure we haven't tried to authenticate before getting
         # the password callback
-        self.assertEqual(self.transport.value(), '')
+        self.assertEqual(self.transport.value(), b'')
         d.callback('foo')
 
         # now make sure we DID try to authenticate
         self.assertEqual(
             self.transport.value(),
-            'AUTHENTICATE %s\r\n' % "foo".encode("hex")
+            b'AUTHENTICATE ' + b2a_hex(b"foo") + b'\r\n'
         )
 
     def test_authenticate_password_deferred_but_no_password(self):
         d = defer.Deferred()
         self.protocol.password_function = lambda: d
         self.protocol.makeConnection(self.transport)
-        self.assertEqual(self.transport.value(), 'PROTOCOLINFO 1\r\n')
+        self.assertEqual(self.transport.value(), b'PROTOCOLINFO 1\r\n')
         self.transport.clear()
-        self.send('250-PROTOCOLINFO 1')
-        self.send('250-AUTH METHODS=HASHEDPASSWORD')
-        self.send('250-VERSION Tor="0.2.2.34"')
-        self.send('250 OK')
+        self.send(b'250-PROTOCOLINFO 1')
+        self.send(b'250-AUTH METHODS=HASHEDPASSWORD')
+        self.send(b'250-VERSION Tor="0.2.2.34"')
+        self.send(b'250 OK')
         d.callback(None)
         return self.assertFailure(self.protocol.post_bootstrap, RuntimeError)
 
@@ -158,12 +178,12 @@ class AuthenticationTests(unittest.TestCase):
         self.auth_failed = False
 
         self.protocol.makeConnection(self.transport)
-        self.assertEqual(self.transport.value(), 'PROTOCOLINFO 1\r\n')
+        self.assertEqual(self.transport.value(), b'PROTOCOLINFO 1\r\n')
 
-        self.send('250-PROTOCOLINFO 1')
-        self.send('250-AUTH METHODS=HASHEDPASSWORD')
-        self.send('250-VERSION Tor="0.2.2.34"')
-        self.send('250 OK')
+        self.send(b'250-PROTOCOLINFO 1')
+        self.send(b'250-AUTH METHODS=HASHEDPASSWORD')
+        self.send(b'250-VERSION Tor="0.2.2.34"')
+        self.send(b'250 OK')
 
         self.assertTrue(self.auth_failed)
 
@@ -223,13 +243,14 @@ class ProtocolTests(unittest.TestCase):
         self.protocol = None
 
     def send(self, line):
-        self.protocol.dataReceived(line.strip() + "\r\n")
+        assert type(line) == bytes
+        self.protocol.dataReceived(line.strip() + b"\r\n")
 
     def test_statemachine_broadcast_no_code(self):
         try:
             self.protocol._broadcast_response("foo")
             self.fail()
-        except RuntimeError, e:
+        except RuntimeError as e:
             self.assertTrue('No code set yet' in str(e))
 
     def test_statemachine_broadcast_unknown_code(self):
@@ -237,7 +258,7 @@ class ProtocolTests(unittest.TestCase):
             self.protocol.code = 999
             self.protocol._broadcast_response("foo")
             self.fail()
-        except RuntimeError, e:
+        except RuntimeError as e:
             self.assertTrue('Unknown code' in str(e))
 
     def test_statemachine_is_finish(self):
@@ -254,7 +275,7 @@ class ProtocolTests(unittest.TestCase):
             self.protocol.code = 250
             self.protocol._is_continuation_line("123 ")
             self.fail()
-        except RuntimeError, e:
+        except RuntimeError as e:
             self.assertTrue('Unexpected code' in str(e))
 
     def test_statemachine_multiline(self):
@@ -262,9 +283,17 @@ class ProtocolTests(unittest.TestCase):
             self.protocol.code = 250
             self.protocol._is_multi_line("123 ")
             self.fail()
-        except RuntimeError, e:
+        except RuntimeError as e:
             self.assertTrue('Unexpected code' in str(e))
 
+    def test_response_with_no_request(self):
+        with self.assertRaises(RuntimeError) as ctx:
+            self.protocol.code = 200
+            self.protocol._broadcast_response('200 OK')
+        self.assertTrue(
+            "didn't issue a command" in str(ctx.exception)
+        )
+
     def auth_failed(self, msg):
         self.assertEqual(str(msg.value), '551 go away')
         self.got_auth_failed = True
@@ -278,7 +307,7 @@ class ProtocolTests(unittest.TestCase):
 AUTH METHODS=HASHEDPASSWORD
 VERSION Tor="0.2.2.35"
 OK''')
-        self.send('551 go away\r\n')
+        self.send(b'551 go away\r\n')
         self.assertTrue(self.got_auth_failed)
 
     def test_authenticate_no_auth_line(self):
@@ -288,12 +317,12 @@ FOOAUTH METHODS=COOKIE,SAFECOOKIE COOKIEFILE="/dev/null"
 VERSION Tor="0.2.2.35"
 OK''')
             self.assertTrue(False)
-        except RuntimeError, e:
+        except RuntimeError as e:
             self.assertTrue('find AUTH line' in str(e))
 
     def test_authenticate_not_enough_cookie_data(self):
         with tempfile.NamedTemporaryFile() as cookietmp:
-            cookietmp.write('x' * 35)  # too much data
+            cookietmp.write(b'x' * 35)  # too much data
             cookietmp.flush()
 
             try:
@@ -302,12 +331,12 @@ AUTH METHODS=COOKIE COOKIEFILE="%s"
 VERSION Tor="0.2.2.35"
 OK''' % cookietmp.name)
                 self.assertTrue(False)
-            except RuntimeError, e:
+            except RuntimeError as e:
                 self.assertTrue('cookie to be 32' in str(e))
 
     def test_authenticate_not_enough_safecookie_data(self):
         with tempfile.NamedTemporaryFile() as cookietmp:
-            cookietmp.write('x' * 35)  # too much data
+            cookietmp.write(b'x' * 35)  # too much data
             cookietmp.flush()
 
             try:
@@ -316,46 +345,48 @@ AUTH METHODS=SAFECOOKIE COOKIEFILE="%s"
 VERSION Tor="0.2.2.35"
 OK''' % cookietmp.name)
                 self.assertTrue(False)
-            except RuntimeError, e:
+            except RuntimeError as e:
                 self.assertTrue('cookie to be 32' in str(e))
 
     def test_authenticate_safecookie(self):
         with tempfile.NamedTemporaryFile() as cookietmp:
-            cookiedata = str(bytearray([0] * 32))
+            cookiedata = bytes(bytearray([0] * 32))
             cookietmp.write(cookiedata)
             cookietmp.flush()
 
             self.protocol._do_authenticate('''PROTOCOLINFO 1
-AUTH METHODS=SAFECOOKIE COOKIEFILE="%s"
+AUTH METHODS=SAFECOOKIE COOKIEFILE="{}"
 VERSION Tor="0.2.2.35"
-OK''' % cookietmp.name)
+OK'''.format(cookietmp.name))
             self.assertTrue(
-                'AUTHCHALLENGE SAFECOOKIE ' in self.transport.value()
+                b'AUTHCHALLENGE SAFECOOKIE ' in self.transport.value()
             )
-            client_nonce = base64.b16decode(self.transport.value().split()[-1])
+            x = self.transport.value().split()[-1]
+            client_nonce = a2b_hex(x)
             self.transport.clear()
-            server_nonce = str(bytearray([0] * 32))
+            server_nonce = bytes(bytearray([0] * 32))
             server_hash = hmac_sha256(
-                "Tor safe cookie authentication server-to-controller hash",
-                cookiedata + client_nonce + server_nonce
+                b"Tor safe cookie authentication server-to-controller hash",
+                cookiedata + client_nonce + server_nonce,
             )
 
             self.send(
-                '250 AUTHCHALLENGE SERVERHASH=%s SERVERNONCE=%s' %
-                (base64.b16encode(server_hash), base64.b16encode(server_nonce))
+                b'250 AUTHCHALLENGE SERVERHASH=' +
+                base64.b16encode(server_hash) + b' SERVERNONCE=' +
+                base64.b16encode(server_nonce) + b'\r\n'
             )
-            self.assertTrue('AUTHENTICATE ' in self.transport.value())
+            self.assertTrue(b'AUTHENTICATE ' in self.transport.value())
 
     def test_authenticate_cookie_without_reading(self):
-        server_nonce = str(bytearray([0] * 32))
-        server_hash = str(bytearray([0] * 32))
+        server_nonce = bytes(bytearray([0] * 32))
+        server_hash = bytes(bytearray([0] * 32))
         try:
             self.protocol._safecookie_authchallenge(
                 '250 AUTHCHALLENGE SERVERHASH=%s SERVERNONCE=%s' %
                 (base64.b16encode(server_hash), base64.b16encode(server_nonce))
             )
             self.assertTrue(False)
-        except RuntimeError, e:
+        except RuntimeError as e:
             self.assertTrue('not read' in str(e))
 
     def test_authenticate_unexisting_cookie_file(self):
@@ -373,9 +404,9 @@ OK''' % unexisting_file)
         unexisting_file = __file__ + "-unexisting"
         try:
             self.protocol._do_authenticate('''PROTOCOLINFO 1
-AUTH METHODS=SAFECOOKIE COOKIEFILE="%s"
+AUTH METHODS=SAFECOOKIE COOKIEFILE="{}"
 VERSION Tor="0.2.2.35"
-OK''' % unexisting_file)
+OK'''.format(unexisting_file))
             self.assertTrue(False)
         except RuntimeError:
             pass
@@ -394,24 +425,30 @@ OK''')
         unexisting_file = __file__ + "-unexisting"
         self.protocol.password_function = lambda: 'foo'
         self.protocol._do_authenticate('''PROTOCOLINFO 1
-AUTH METHODS=COOKIE,HASHEDPASSWORD COOKIEFILE="%s"
+AUTH METHODS=COOKIE,HASHEDPASSWORD COOKIEFILE="{}"
 VERSION Tor="0.2.2.35"
-OK''' % unexisting_file)
-        self.assertEqual(self.transport.value(), 'AUTHENTICATE %s\r\n' % "foo".encode("hex"))
+OK'''.format(unexisting_file))
+        self.assertEqual(
+            self.transport.value(),
+            b'AUTHENTICATE ' + b2a_hex(b'foo') + b'\r\n',
+        )
 
     def test_authenticate_password_when_safecookie_unavailable(self):
         unexisting_file = __file__ + "-unexisting"
         self.protocol.password_function = lambda: 'foo'
         self.protocol._do_authenticate('''PROTOCOLINFO 1
-AUTH METHODS=SAFECOOKIE,HASHEDPASSWORD COOKIEFILE="%s"
+AUTH METHODS=SAFECOOKIE,HASHEDPASSWORD COOKIEFILE="{}"
 VERSION Tor="0.2.2.35"
-OK''' % unexisting_file)
-        self.assertEqual(self.transport.value(), 'AUTHENTICATE %s\r\n' % "foo".encode("hex"))
+OK'''.format(unexisting_file))
+        self.assertEqual(
+            self.transport.value(),
+            b'AUTHENTICATE ' + b2a_hex(b'foo') + b'\r\n',
+        )
 
     def test_authenticate_safecookie_wrong_hash(self):
-        cookiedata = str(bytearray([0] * 32))
-        server_nonce = str(bytearray([0] * 32))
-        server_hash = str(bytearray([0] * 32))
+        cookiedata = bytes(bytearray([0] * 32))
+        server_nonce = bytes(bytearray([0] * 32))
+        server_hash = bytes(bytearray([0] * 32))
 
         # pretend we already did PROTOCOLINFO and read the cookie
         # file
@@ -419,11 +456,13 @@ OK''' % unexisting_file)
         self.protocol.client_nonce = server_nonce  # all 0's anyway
         try:
             self.protocol._safecookie_authchallenge(
-                '250 AUTHCHALLENGE SERVERHASH=%s SERVERNONCE=%s' %
-                (base64.b16encode(server_hash), base64.b16encode(server_nonce))
+                '250 AUTHCHALLENGE SERVERHASH={} SERVERNONCE={}'.format(
+                    b2a_hex(server_hash).decode('ascii'),
+                    b2a_hex(server_nonce).decode('ascii'),
+                )
             )
             self.assertTrue(False)
-        except RuntimeError, e:
+        except RuntimeError as e:
             self.assertTrue('hash not expected' in str(e))
 
     def confirm_version_events(self, arg):
@@ -437,26 +476,26 @@ OK''' % unexisting_file)
         d.addCallback(CallbackChecker(self.protocol))
         d.addCallback(self.confirm_version_events)
 
-        events = 'GUARD STREAM CIRC NS NEWCONSENSUS ORCONN NEWDESC ADDRMAP STATUS_GENERAL'
+        events = b'GUARD STREAM CIRC NS NEWCONSENSUS ORCONN NEWDESC ADDRMAP STATUS_GENERAL'
         self.protocol._bootstrap()
 
         # answer all the requests generated by boostrapping etc.
-        self.send("250-signal/names=")
-        self.send("250 OK")
+        self.send(b"250-signal/names=")
+        self.send(b"250 OK")
 
-        self.send("250-version=foo")
-        self.send("250 OK")
+        self.send(b"250-version=foo")
+        self.send(b"250 OK")
 
-        self.send("250-events/names=" + events)
-        self.send("250 OK")
+        self.send(b"250-events/names=" + events)
+        self.send(b"250 OK")
 
-        self.send("250 OK")  # for USEFEATURE
+        self.send(b"250 OK")  # for USEFEATURE
 
         return d
 
     def test_bootstrap_tor_does_not_support_signal_names(self):
         self.protocol._bootstrap()
-        self.send('552 Unrecognized key "signal/names"')
+        self.send(b'552 Unrecognized key "signal/names"')
         valid_signals = ["RELOAD", "DUMP", "DEBUG", "NEWNYM", "CLEARDNSCACHE"]
         self.assertEqual(self.protocol.valid_signals, valid_signals)
 
@@ -467,12 +506,12 @@ OK''' % unexisting_file)
         """
         self.protocol._set_valid_events('CIRC')
         self.protocol.add_event_listener('CIRC', lambda _: None)
-        self.send("250 OK")
+        self.send(b"250 OK")
 
         d = self.protocol.get_conf("SOCKSPORT ORPORT")
-        self.send("650 CIRC 1000 EXTENDED moria1,moria2")
-        self.send("250-SOCKSPORT=9050")
-        self.send("250 ORPORT=0")
+        self.send(b"650 CIRC 1000 EXTENDED moria1,moria2")
+        self.send(b"250-SOCKSPORT=9050")
+        self.send(b"250 ORPORT=0")
         return d
 
     def test_async_multiline(self):
@@ -489,15 +528,15 @@ OK''' % unexisting_file)
                 "1000 EXTENDED moria1,moria2\nEXTRAMAGIC=99\nANONYMITY=high"
             )
         )
-        self.send("250 OK")
+        self.send(b"250 OK")
 
         d = self.protocol.get_conf("SOCKSPORT ORPORT")
         d.addCallback(CallbackChecker({"ORPORT": "0", "SOCKSPORT": "9050"}))
-        self.send("650-CIRC 1000 EXTENDED moria1,moria2")
-        self.send("650-EXTRAMAGIC=99")
-        self.send("650 ANONYMITY=high")
-        self.send("250-SOCKSPORT=9050")
-        self.send("250 ORPORT=0")
+        self.send(b"650-CIRC 1000 EXTENDED moria1,moria2")
+        self.send(b"650-EXTRAMAGIC=99")
+        self.send(b"650 ANONYMITY=high")
+        self.send(b"250-SOCKSPORT=9050")
+        self.send(b"250 ORPORT=0")
         return d
 
     def test_multiline_plus(self):
@@ -506,12 +545,12 @@ OK''' % unexisting_file)
 
         d = self.protocol.get_info("FOO")
         d.addCallback(CallbackChecker({"FOO": "\na\nb\nc"}))
-        self.send("250+FOO=")
-        self.send("a")
-        self.send("b")
-        self.send("c")
-        self.send(".")
-        self.send("250 OK")
+        self.send(b"250+FOO=")
+        self.send(b"a")
+        self.send(b"b")
+        self.send(b"c")
+        self.send(b".")
+        self.send(b"250 OK")
         return d
 
     def test_multiline_plus_embedded_equals(self):
@@ -520,10 +559,10 @@ OK''' % unexisting_file)
 
         d = self.protocol.get_info("FOO")
         d.addCallback(CallbackChecker({"FOO": "\na="}))
-        self.send("250+FOO=")
-        self.send("a=")
-        self.send(".")
-        self.send("250 OK")
+        self.send(b"250+FOO=")
+        self.send(b"a=")
+        self.send(b".")
+        self.send(b"250 OK")
         return d
 
     def incremental_check(self, expected, actual):
@@ -536,11 +575,11 @@ OK''' % unexisting_file)
             "FOO",
             functools.partial(self.incremental_check, "bar")
         )
-        self.send("250+FOO=")
-        self.send("bar")
-        self.send("bar")
-        self.send(".")
-        self.send("250 OK")
+        self.send(b"250+FOO=")
+        self.send(b"bar")
+        self.send(b"bar")
+        self.send(b".")
+        self.send(b"250 OK")
         return d
 
     def test_getinfo_incremental_continuation(self):
@@ -548,32 +587,32 @@ OK''' % unexisting_file)
             "FOO",
             functools.partial(self.incremental_check, "bar")
         )
-        self.send("250-FOO=")
-        self.send("250-bar")
-        self.send("250-bar")
-        self.send("250 OK")
+        self.send(b"250-FOO=")
+        self.send(b"250-bar")
+        self.send(b"250-bar")
+        self.send(b"250 OK")
         return d
 
     def test_getinfo_one_line(self):
         d = self.protocol.get_info(
             "foo",
-            functools.partial(self.incremental_check, "bar")
         )
-        self.send('250 foo=bar')
+        self.send(b'250 foo=bar')
+        d.addCallback(lambda _: functools.partial(self.incremental_check, "bar"))
         return d
 
     def test_getconf(self):
         d = self.protocol.get_conf("SOCKSPORT ORPORT")
         d.addCallback(CallbackChecker({'SocksPort': '9050', 'ORPort': '0'}))
-        self.send("250-SocksPort=9050")
-        self.send("250 ORPort=0")
+        self.send(b"250-SocksPort=9050")
+        self.send(b"250 ORPort=0")
         return d
 
     def test_getconf_raw(self):
         d = self.protocol.get_conf_raw("SOCKSPORT ORPORT")
         d.addCallback(CallbackChecker('SocksPort=9050\nORPort=0'))
-        self.send("250-SocksPort=9050")
-        self.send("250 ORPort=0")
+        self.send(b"250-SocksPort=9050")
+        self.send(b"250 ORPort=0")
         return d
 
     def response_ok(self, v):
@@ -583,31 +622,37 @@ OK''' % unexisting_file)
         d = self.protocol.set_conf("foo", "bar").addCallback(
             functools.partial(self.response_ok)
         )
-        self.send("250 OK")
+        self.send(b"250 OK")
         self._wait(d)
-        self.assertEqual(self.transport.value(), "SETCONF foo=bar\r\n")
+        self.assertEqual(self.transport.value(), b"SETCONF foo=bar\r\n")
 
     def test_setconf_with_space(self):
         d = self.protocol.set_conf("foo", "a value with a space")
         d.addCallback(functools.partial(self.response_ok))
-        self.send("250 OK")
+        self.send(b"250 OK")
         self._wait(d)
         self.assertEqual(
             self.transport.value(),
-            'SETCONF foo="a value with a space"\r\n'
+            b'SETCONF foo="a value with a space"\r\n'
         )
 
     def test_setconf_multi(self):
         d = self.protocol.set_conf("foo", "bar", "baz", 1)
-        self.send("250 OK")
+        self.send(b"250 OK")
         self._wait(d)
-        self.assertEqual(self.transport.value(), "SETCONF foo=bar baz=1\r\n")
+        self.assertEqual(
+            self.transport.value(),
+            b"SETCONF foo=bar baz=1\r\n",
+        )
 
     def test_quit(self):
         d = self.protocol.quit()
-        self.send("250 OK")
+        self.send(b"250 OK")
         self._wait(d)
-        self.assertEqual(self.transport.value(), "QUIT\r\n")
+        self.assertEqual(
+            self.transport.value(),
+            b"QUIT\r\n",
+        )
 
     def test_dot(self):
         # just checking we don't expode
@@ -618,7 +663,7 @@ OK''' % unexisting_file)
         self.assertTrue(exists('txtorcon-debug.log'))
 
     def error(self, failure):
-        print "ERROR", failure
+        print("ERROR", failure)
         self.assertTrue(False)
 
     def test_twocommands(self):
@@ -630,10 +675,10 @@ OK''' % unexisting_file)
         d2 = self.protocol.get_info_raw("BAR")
         d2.addCallback(CallbackChecker("bar")).addErrback(log.err)
 
-        self.send("250-a=one")
-        self.send("250-b=two")
-        self.send("250 OK")
-        self.send("250 bar")
+        self.send(b"250-a=one")
+        self.send(b"250-b=two")
+        self.send(b"250 OK")
+        self.send(b"250 bar")
 
         return d2
 
@@ -641,13 +686,16 @@ OK''' % unexisting_file)
         try:
             self.protocol.signal('FOO')
             self.fail()
-        except Exception, e:
+        except Exception as e:
             self.assertTrue('Invalid signal' in str(e))
 
     def test_signal(self):
         self.protocol.valid_signals = ['NEWNYM']
         self.protocol.signal('NEWNYM')
-        self.assertEqual(self.transport.value(), 'SIGNAL NEWNYM\r\n')
+        self.assertEqual(
+            self.transport.value(),
+            b'SIGNAL NEWNYM\r\n',
+        )
 
     def test_650_after_authenticate(self):
         self.protocol._set_valid_events('CONF_CHANGED')
@@ -655,10 +703,10 @@ OK''' % unexisting_file)
             'CONF_CHANGED',
             CallbackChecker("Foo=bar")
         )
-        self.send("250 OK")
+        self.send(b"250 OK")
 
-        self.send("650-CONF_CHANGED")
-        self.send("650-Foo=bar")
+        self.send(b"650-CONF_CHANGED")
+        self.send(b"650-Foo=bar")
 
     def test_notify_after_getinfo(self):
         self.protocol._set_valid_events('CIRC')
@@ -666,32 +714,35 @@ OK''' % unexisting_file)
             'CIRC',
             CallbackChecker("1000 EXTENDED moria1,moria2")
         )
-        self.send("250 OK")
+        self.send(b"250 OK")
 
         d = self.protocol.get_info("a")
         d.addCallback(CallbackChecker({'a': 'one'})).addErrback(self.fail)
-        self.send("250-a=one")
-        self.send("250 OK")
-        self.send("650 CIRC 1000 EXTENDED moria1,moria2")
+        self.send(b"250-a=one")
+        self.send(b"250 OK")
+        self.send(b"650 CIRC 1000 EXTENDED moria1,moria2")
         return d
 
     def test_notify_error(self):
         self.protocol._set_valid_events('CIRC')
-        self.send("650 CIRC 1000 EXTENDED moria1,moria2")
+        self.send(b"650 CIRC 1000 EXTENDED moria1,moria2")
 
     def test_getinfo(self):
         d = self.protocol.get_info("version")
         d.addCallback(CallbackChecker({'version': '0.2.2.34'}))
         d.addErrback(self.fail)
 
-        self.send("250-version=0.2.2.34")
-        self.send("250 OK")
+        self.send(b"250-version=0.2.2.34")
+        self.send(b"250 OK")
 
-        self.assertEqual(self.transport.value(), "GETINFO version\r\n")
+        self.assertEqual(
+            self.transport.value(),
+            b"GETINFO version\r\n",
+        )
         return d
 
     def test_getinfo_for_descriptor(self):
-        descriptor_info = """250+desc/name/moria1=
+        descriptor_info = b"""250+desc/name/moria1=
 router moria1 128.31.0.34 9101 0 9131
 platform Tor 0.2.5.0-alpha-dev on Linux
 protocols Link 1 2 Circuit 1
@@ -726,15 +777,15 @@ iO3EUE0AEYah2W9gdz8t+i3Dtr0zgqLS841GC/TyDKCm+MKmN8d098qnwK0NGF9q
 .
 250 OK"""
         d = self.protocol.get_info("desc/name/moria1")
-        d.addCallback(CallbackChecker({'desc/name/moria1': '\n' + '\n'.join(descriptor_info.split('\n')[1:-2])}))
+        d.addCallback(CallbackChecker({'desc/name/moria1': '\n' + '\n'.join(descriptor_info.decode('ascii').split('\n')[1:-2])}))
         d.addErrback(self.fail)
 
-        for line in descriptor_info.split('\n'):
+        for line in descriptor_info.split(b'\n'):
             self.send(line)
         return d
 
     def test_getinfo_multiline(self):
-        descriptor_info = """250+desc/name/moria1=
+        descriptor_info = b"""250+desc/name/moria1=
 router moria1 128.31.0.34 9101 0 9131
 platform Tor 0.2.5.0-alpha-dev on Linux
 .
@@ -744,7 +795,7 @@ platform Tor 0.2.5.0-alpha-dev on Linux
         d.addCallback(CallbackChecker({'desc/name/moria1': gold}))
         d.addErrback(self.fail)
 
-        for line in descriptor_info.split('\n'):
+        for line in descriptor_info.split(b'\n'):
             self.send(line)
         return d
 
@@ -755,19 +806,21 @@ platform Tor 0.2.5.0-alpha-dev on Linux
         # is it dangerous/ill-advised to depend on internal state of
         # class under test?
         d = self.protocol.defer
-        self.send("250 OK")
+        self.send(b"250 OK")
         self._wait(d)
         self.assertEqual(
-            self.transport.value().split('\r\n')[-2],
-            "SETEVENTS FOO"
+            self.transport.value().split(b'\r\n')[-2],
+            b"SETEVENTS FOO"
         )
         self.transport.clear()
 
         self.protocol.add_event_listener('BAR', lambda _: None)
         d = self.protocol.defer
-        self.send("250 OK")
-        self.assertTrue(self.transport.value() == "SETEVENTS FOO BAR\r\n" or
-                        self.transport.value() == "SETEVENTS BAR FOO\r\n")
+        self.send(b"250 OK")
+        self.assertTrue(
+            self.transport.value() == b"SETEVENTS FOO BAR\r\n" or
+            self.transport.value() == b"SETEVENTS BAR FOO\r\n"
+        )
         self._wait(d)
 
         try:
@@ -791,12 +844,46 @@ platform Tor 0.2.5.0-alpha-dev on Linux
         self.protocol.add_event_listener('STREAM', listener)
 
         d = self.protocol.defer
-        self.send("250 OK")
+        self.send(b"250 OK")
         self._wait(d)
-        self.send("650 STREAM 1234 NEW 4321 1.2.3.4:555 REASON=MISC")
-        self.send("650 STREAM 2345 NEW 4321 2.3.4.5:666 REASON=MISC")
+        self.send(b"650 STREAM 1234 NEW 4321 1.2.3.4:555 REASON=MISC")
+        self.send(b"650 STREAM 2345 NEW 4321 2.3.4.5:666 REASON=MISC")
         self.assertEqual(listener.stream_events, 2)
 
+    def test_eventlistener_error(self):
+        self.protocol._set_valid_events('STREAM')
+
+        class EventListener(object):
+            stream_events = 0
+            do_error = False
+
+            def __call__(self, data):
+                self.stream_events += 1
+                if self.do_error:
+                    raise Exception("the bad thing happened")
+
+        # we make sure the first listener has the errors to prove the
+        # second one still gets called.
+        listener0 = EventListener()
+        listener0.do_error = True
+        listener1 = EventListener()
+        self.protocol.add_event_listener('STREAM', listener0)
+        self.protocol.add_event_listener('STREAM', listener1)
+
+        d = self.protocol.defer
+        self.send(b"250 OK")
+        self._wait(d)
+        self.send(b"650 STREAM 1234 NEW 4321 1.2.3.4:555 REASON=MISC")
+        self.send(b"650 STREAM 2345 NEW 4321 2.3.4.5:666 REASON=MISC")
+        self.assertEqual(listener0.stream_events, 2)
+        self.assertEqual(listener1.stream_events, 2)
+
+        # should have logged the two errors
+        logged = self.flushLoggedErrors()
+        self.assertEqual(2, len(logged))
+        self.assertTrue("the bad thing happened" in str(logged[0]))
+        self.assertTrue("the bad thing happened" in str(logged[1]))
+
     def test_remove_eventlistener(self):
         self.protocol._set_valid_events('STREAM')
 
@@ -808,11 +895,11 @@ platform Tor 0.2.5.0-alpha-dev on Linux
 
         listener = EventListener()
         self.protocol.add_event_listener('STREAM', listener)
-        self.assertEqual(self.transport.value(), 'SETEVENTS STREAM\r\n')
-        self.protocol.lineReceived("250 OK")
+        self.assertEqual(self.transport.value(), b'SETEVENTS STREAM\r\n')
+        self.protocol.lineReceived(b"250 OK")
         self.transport.clear()
         self.protocol.remove_event_listener('STREAM', listener)
-        self.assertEqual(self.transport.value(), 'SETEVENTS \r\n')
+        self.assertEqual(self.transport.value(), b'SETEVENTS \r\n')
 
     def test_remove_eventlistener_multiple(self):
         self.protocol._set_valid_events('STREAM')
@@ -826,41 +913,40 @@ platform Tor 0.2.5.0-alpha-dev on Linux
         listener0 = EventListener()
         listener1 = EventListener()
         self.protocol.add_event_listener('STREAM', listener0)
-        self.assertEqual(self.transport.value(), 'SETEVENTS STREAM\r\n')
-        self.protocol.lineReceived("250 OK")
+        self.assertEqual(self.transport.value(), b'SETEVENTS STREAM\r\n')
+        self.protocol.lineReceived(b"250 OK")
         self.transport.clear()
         # add another one, shouldn't issue a tor command
         self.protocol.add_event_listener('STREAM', listener1)
-        self.assertEqual(self.transport.value(), '')
+        self.assertEqual(self.transport.value(), b'')
 
         # remove one, should still not issue a tor command
         self.protocol.remove_event_listener('STREAM', listener0)
-        self.assertEqual(self.transport.value(), '')
+        self.assertEqual(self.transport.value(), b'')
 
         # remove the other one, NOW should issue a command
         self.protocol.remove_event_listener('STREAM', listener1)
-        self.assertEqual(self.transport.value(), 'SETEVENTS \r\n')
+        self.assertEqual(self.transport.value(), b'SETEVENTS \r\n')
 
         # try removing invalid event
         try:
             self.protocol.remove_event_listener('FOO', listener0)
             self.fail()
-        except Exception, e:
+        except Exception as e:
             self.assertTrue('FOO' in str(e))
 
-    def checkContinuation(self, v):
-        self.assertEqual(v, "key=\nvalue0\nvalue1")
-
-    def test_continuationLine(self):
+    def test_continuation_line(self):
         d = self.protocol.get_info_raw("key")
 
-        d.addCallback(self.checkContinuation)
+        def check_continuation(v):
+            self.assertEqual(v, "key=\nvalue0\nvalue1")
+        d.addCallback(check_continuation)
 
-        self.send("250+key=")
-        self.send("value0")
-        self.send("value1")
-        self.send(".")
-        self.send("250 OK")
+        self.send(b"250+key=")
+        self.send(b"value0")
+        self.send(b"value1")
+        self.send(b".")
+        self.send(b"250 OK")
 
         return d
 
@@ -879,19 +965,19 @@ s Exit Fast Guard HSDir Named Running Stable V2Dir Valid
 w Bandwidth=518000
 p accept 43,53,79-81,110,143,194,220,443,953,989-990,993,995,1194,1293,1723,1863,2082-2083,2086-2087,2095-2096,3128,4321,5050,5190,5222-5223,6679,6697,7771,8000,8008,8080-8081,8090,8118,8123,8181,8300,8443,8888"""))
 
-        self.send("250+ns/id/624926802351575FF7E4E3D60EFA3BFB56E67E8A=")
-        self.send("r fake YkkmgCNRV1/35OPWDvo7+1bmfoo tanLV/4ZfzpYQW0xtGFqAa46foo 2011-12-12 16:29:16 12.45.56.78 443 80")
-        self.send("s Exit Fast Guard HSDir Named Running Stable V2Dir Valid")
-        self.send("w Bandwidth=518000")
-        self.send("p accept 43,53,79-81,110,143,194,220,443,953,989-990,993,995,1194,1293,1723,1863,2082-2083,2086-2087,2095-2096,3128,4321,5050,5190,5222-5223,6679,6697,7771,8000,8008,8080-8081,8090,8118,8123,8181,8300,8443,8888")
-        self.send(".")
-        self.send("250 OK")
+        self.send(b"250+ns/id/624926802351575FF7E4E3D60EFA3BFB56E67E8A=")
+        self.send(b"r fake YkkmgCNRV1/35OPWDvo7+1bmfoo tanLV/4ZfzpYQW0xtGFqAa46foo 2011-12-12 16:29:16 12.45.56.78 443 80")
+        self.send(b"s Exit Fast Guard HSDir Named Running Stable V2Dir Valid")
+        self.send(b"w Bandwidth=518000")
+        self.send(b"p accept 43,53,79-81,110,143,194,220,443,953,989-990,993,995,1194,1293,1723,1863,2082-2083,2086-2087,2095-2096,3128,4321,5050,5190,5222-5223,6679,6697,7771,8000,8008,8080-8081,8090,8118,8123,8181,8300,8443,8888")
+        self.send(b".")
+        self.send(b"250 OK")
 
         return d
 
     def test_plus_line_no_command(self):
-        self.protocol.lineReceived("650+NS\r\n")
-        self.protocol.lineReceived("r Gabor gFpAHsFOHGATy12ZUswRf0ZrqAU GG6GDp40cQfR3ODvkBT0r+Q09kw 2012-05-12 16:54:56 91.219.238.71 443 80\r\n")
+        self.protocol.lineReceived(b"650+NS\r\n")
+        self.protocol.lineReceived(b"r Gabor gFpAHsFOHGATy12ZUswRf0ZrqAU GG6GDp40cQfR3ODvkBT0r+Q09kw 2012-05-12 16:54:56 91.219.238.71 443 80\r\n")
 
     def test_minus_line_no_command(self):
         """
@@ -900,8 +986,8 @@ p accept 43,53,79-81,110,143,194,220,443,953,989-990,993,995,1194,1293,1723,1863
         """
         self.protocol._set_valid_events('NS')
         self.protocol.add_event_listener('NS', lambda _: None)
-        self.protocol.lineReceived("650-NS\r\n")
-        self.protocol.lineReceived("650 OK\r\n")
+        self.protocol.lineReceived(b"650-NS\r\n")
+        self.protocol.lineReceived(b"650 OK\r\n")
 
 
 class ParseTests(unittest.TestCase):
@@ -931,7 +1017,7 @@ class ParseTests(unittest.TestCase):
     def test_multientry_keywords_2(self):
         x = parse_keywords('foo=bar\nfoo=zarimba')
         self.assertEqual(len(x), 1)
-        self.assertTrue(isinstance(x['foo'], types.ListType))
+        self.assertTrue(isinstance(x['foo'], list))
         self.assertEqual(len(x['foo']), 2)
         self.assertEqual(x['foo'][0], 'bar')
         self.assertEqual(x['foo'][1], 'zarimba')
@@ -939,7 +1025,7 @@ class ParseTests(unittest.TestCase):
     def test_multientry_keywords_3(self):
         x = parse_keywords('foo=bar\nfoo=baz\nfoo=zarimba')
         self.assertEqual(len(x), 1)
-        self.assertTrue(isinstance(x['foo'], types.ListType))
+        self.assertTrue(isinstance(x['foo'], list))
         self.assertEqual(len(x['foo']), 3)
         self.assertEqual(x['foo'][0], 'bar')
         self.assertEqual(x['foo'][1], 'baz')
@@ -948,7 +1034,7 @@ class ParseTests(unittest.TestCase):
     def test_multientry_keywords_4(self):
         x = parse_keywords('foo=bar\nfoo=baz\nfoo=zarimba\nfoo=foo')
         self.assertEqual(len(x), 1)
-        self.assertTrue(isinstance(x['foo'], types.ListType))
+        self.assertTrue(isinstance(x['foo'], list))
         self.assertEqual(len(x['foo']), 4)
         self.assertEqual(x['foo'][0], 'bar')
         self.assertEqual(x['foo'][1], 'baz')
@@ -967,8 +1053,7 @@ w Bandwidth=1234
 OK
 ''')
         self.assertEqual(2, len(x))
-        keys = x.keys()
-        keys.sort()
+        keys = sorted(x.keys())
         self.assertEqual(keys, ['ns/name/bar', 'ns/name/foo'])
 
     def test_multiline_keywords(self):
