@@ -1,4 +1,6 @@
-from zope.interface import implements
+from __future__ import print_function
+
+from zope.interface import implementer, directlyProvides
 from zope.interface.verify import verifyClass
 from twisted.trial import unittest
 from twisted.test import proto_helpers
@@ -6,10 +8,12 @@ from twisted.python.failure import Failure
 from twisted.internet import task, defer
 from twisted.internet.interfaces import IStreamClientEndpoint, IReactorCore
 
-import os
 import tempfile
+import six
 
-from mock import patch
+from ipaddress import IPv4Address
+
+from mock import patch, Mock
 
 from txtorcon import TorControlProtocol
 from txtorcon import TorProtocolError
@@ -20,16 +24,20 @@ from txtorcon import build_tor_connection
 from txtorcon import build_local_tor_connection
 from txtorcon import build_timeout_circuit
 from txtorcon import CircuitBuildTimedOutError
-from txtorcon.interface import ITorControlProtocol
 from txtorcon.interface import IStreamAttacher
 from txtorcon.interface import ICircuitListener
 from txtorcon.interface import IStreamListener
 from txtorcon.interface import StreamListenerMixin
 from txtorcon.interface import CircuitListenerMixin
+from txtorcon.torstate import _extract_reason
+from txtorcon.circuit import _get_circuit_attacher
+
+if six.PY3:
+    from .py3_torstate import TorStatePy3Tests  # noqa
 
 
+@implementer(ICircuitListener)
 class CircuitListener(object):
-    implements(ICircuitListener)
 
     def __init__(self, expected):
         "expect is a list of tuples: (event, {key:value, key1:value1, ..})"
@@ -73,8 +81,8 @@ class CircuitListener(object):
         self.checker('failed', circuit, **kw)
 
 
+@implementer(IStreamListener)
 class StreamListener(object):
-    implements(IStreamListener)
 
     def __init__(self, expected):
         "expect is a list of tuples: (event, {key:value, key1:value1, ..})"
@@ -115,8 +123,8 @@ class StreamListener(object):
         self.checker('failed', stream, reason)
 
 
+@implementer(IReactorCore)
 class FakeReactor:
-    implements(IReactorCore)
 
     def __init__(self, test):
         self.test = test
@@ -147,8 +155,8 @@ class FakeCircuit(Circuit):
         self.state = 'BOGUS'
 
 
+@implementer(IStreamClientEndpoint)
 class FakeEndpoint:
-    implements(IStreamClientEndpoint)
 
     def get_info_raw(self, keys):
         ans = '\r\n'.join(map(lambda k: '%s=' % k, keys.split()))
@@ -170,8 +178,8 @@ class FakeEndpoint:
         return defer.succeed(self.proto)
 
 
+@implementer(IStreamClientEndpoint)
 class FakeEndpointAnswers:
-    implements(IStreamClientEndpoint)
 
     def __init__(self, answers):
         self.answers = answers
@@ -187,7 +195,13 @@ class FakeEndpointAnswers:
         return ans[:-2]                 # don't want trailing \r\n
 
     def get_info_incremental(self, key, linecb):
-        linecb('%s=%s' % (key, self.answers.pop()))
+        data = self.answers.pop().split('\n')
+        if len(data) == 1:
+            linecb('{}={}'.format(key, data[0]))
+        else:
+            linecb('{}='.format(key))
+            for line in data:
+                linecb(line)
         return defer.succeed('')
 
     def connect(self, protocol_factory):
@@ -207,12 +221,18 @@ class BootstrapTests(unittest.TestCase):
     def confirm_proto(self, x):
         self.assertTrue(isinstance(x, TorControlProtocol))
         self.assertTrue(x.post_bootstrap.called)
+        return x
 
     def confirm_state(self, x):
-        self.assertTrue(isinstance(x, TorState))
+        self.assertIsInstance(x, TorState)
         self.assertTrue(x.post_bootstrap.called)
         return x
 
+    def confirm_consensus(self, x):
+        self.assertEqual(1, len(x.all_routers))
+        self.assertEqual('fake', list(x.routers.values())[0].name)
+        return x
+
     def test_build(self):
         p = FakeEndpoint()
         d = build_tor_connection(p, build_state=False)
@@ -252,6 +272,7 @@ class BootstrapTests(unittest.TestCase):
 
     def confirm_pid(self, state):
         self.assertEqual(state.tor_pid, 1234)
+        return state
 
     def confirm_no_pid(self, state):
         self.assertEqual(state.tor_pid, 0)
@@ -271,6 +292,29 @@ class BootstrapTests(unittest.TestCase):
         p.proto.post_bootstrap.callback(p.proto)
         return d
 
+    def test_build_with_answers_ns(self):
+        fake_consensus = '\n'.join([
+            'r fake YkkmgCNRV1/35OPWDvo7+1bmfoo tanLV/4ZfzpYQW0xtGFqAa46foo 2011-12-12 16:29:16 11.11.11.11 443 80',
+            's Exit Fast Guard HSDir Named Running Stable V2Dir Valid FutureProof',
+            'r ekaf foooooooooooooooooooooooooo barbarbarbarbarbarbarbarbar 2011-11-11 16:30:00 22.22.22.22 443 80',
+            's Exit Fast Guard HSDir Named Running Stable V2Dir Valid FutureProof',
+            '',
+        ])
+        p = FakeEndpointAnswers([fake_consensus,     # ns/all
+                                 '',     # circuit-status
+                                 '',     # stream-status
+                                 '',     # address-mappings/all
+                                 '',     # entry-guards
+                                 '1234'  # PID
+                                 ])
+
+        d = build_tor_connection(p, build_state=True)
+        d.addCallback(self.confirm_state).addErrback(self.fail)
+        d.addCallback(self.confirm_pid).addErrback(self.fail)
+        d.addCallback(self.confirm_consensus).addErrback(self.fail)
+        p.proto.post_bootstrap.callback(p.proto)
+        return d
+
     def test_build_with_answers_no_pid(self):
         p = FakeEndpointAnswers(['',    # ns/all
                                  '',    # circuit-status
@@ -312,6 +356,13 @@ class BootstrapTests(unittest.TestCase):
         return d
 
 
+class UtilTests(unittest.TestCase):
+
+    def test_extract_reason_no_reason(self):
+        reason = _extract_reason(dict())
+        self.assertEqual("unknown", reason)
+
+
 class StateTests(unittest.TestCase):
 
     def setUp(self):
@@ -324,8 +375,8 @@ class StateTests(unittest.TestCase):
         self.protocol.makeConnection(self.transport)
 
     def test_close_stream_with_attacher(self):
+        @implementer(IStreamAttacher)
         class MyAttacher(object):
-            implements(IStreamAttacher)
 
             def __init__(self):
                 self.streams = []
@@ -340,7 +391,7 @@ class StateTests(unittest.TestCase):
 
     def test_attacher_error_handler(self):
         # make sure error-handling "does something" that isn't blowing up
-        with patch('sys.stdout') as fake_stdout:
+        with patch('sys.stdout'):
             TorState(self.protocol)._attacher_error(Failure(RuntimeError("quote")))
 
     def test_stream_update(self):
@@ -369,7 +420,7 @@ class StateTests(unittest.TestCase):
         self.assertEqual(len(self.state.streams), 2)
 
     def send(self, line):
-        self.protocol.dataReceived(line.strip() + "\r\n")
+        self.protocol.dataReceived(line.strip() + b"\r\n")
 
     @defer.inlineCallbacks
     def test_bootstrap_callback(self):
@@ -378,7 +429,7 @@ class StateTests(unittest.TestCase):
         exception from TorState.bootstrap and we'll just hang...
         '''
 
-        from test_torconfig import FakeControlProtocol
+        from .test_torconfig import FakeControlProtocol
         protocol = FakeControlProtocol(
             [
                 "ns/all=",  # ns/all
@@ -413,39 +464,39 @@ class StateTests(unittest.TestCase):
         self.protocol._set_valid_events(' '.join(self.state.event_map.keys()))
         self.state._bootstrap()
 
-        self.send("250+ns/all=")
-        self.send(".")
-        self.send("250 OK")
+        self.send(b"250+ns/all=")
+        self.send(b".")
+        self.send(b"250 OK")
 
-        self.send("250+circuit-status=")
-        self.send(".")
-        self.send("250 OK")
+        self.send(b"250+circuit-status=")
+        self.send(b".")
+        self.send(b"250 OK")
 
-        self.send("250-stream-status=")
-        self.send("250 OK")
+        self.send(b"250-stream-status=")
+        self.send(b"250 OK")
 
-        self.send("250+address-mappings/all=")
-        self.send('www.example.com 127.0.0.1 "2012-01-01 00:00:00"')
-        self.send('subdomain.example.com 10.0.0.0 "2012-01-01 00:01:02"')
-        self.send('.')
-        self.send('250 OK')
+        self.send(b"250+address-mappings/all=")
+        self.send(b'www.example.com 127.0.0.1 "2012-01-01 00:00:00"')
+        self.send(b'subdomain.example.com 10.0.0.0 "2012-01-01 00:01:02"')
+        self.send(b".")
+        self.send(b"250 OK")
 
         for ignored in self.state.event_map.items():
-            self.send("250 OK")
+            self.send(b"250 OK")
 
-        self.send("250-entry-guards=")
-        self.send("250 OK")
+        self.send(b"250-entry-guards=")
+        self.send(b"250 OK")
 
-        self.send("250 OK")
+        self.send(b"250 OK")
 
         self.assertEqual(len(self.state.addrmap.addr), 4)
         self.assertTrue('www.example.com' in self.state.addrmap.addr)
         self.assertTrue('subdomain.example.com' in self.state.addrmap.addr)
         self.assertTrue('10.0.0.0' in self.state.addrmap.addr)
         self.assertTrue('127.0.0.1' in self.state.addrmap.addr)
-        self.assertEqual('127.0.0.1', self.state.addrmap.find('www.example.com').ip)
+        self.assertEqual(IPv4Address(u'127.0.0.1'), self.state.addrmap.find('www.example.com').ip)
         self.assertEqual('www.example.com', self.state.addrmap.find('127.0.0.1').name)
-        self.assertEqual('10.0.0.0', self.state.addrmap.find('subdomain.example.com').ip)
+        self.assertEqual(IPv4Address(u'10.0.0.0'), self.state.addrmap.find('subdomain.example.com').ip)
         self.assertEqual('subdomain.example.com', self.state.addrmap.find('10.0.0.0').name)
 
         return d
@@ -464,55 +515,119 @@ class StateTests(unittest.TestCase):
         self.protocol._set_valid_events(' '.join(self.state.event_map.keys()))
         self.state._bootstrap()
 
-        self.send("250+ns/all=")
-        self.send(".")
-        self.send("250 OK")
+        self.send(b"250+ns/all=")
+        self.send(b".")
+        self.send(b"250 OK")
 
-        self.send("250-circuit-status=123 BUILT PURPOSE=GENERAL")
-        self.send("250 OK")
+        self.send(b"250-circuit-status=123 BUILT PURPOSE=GENERAL")
+        self.send(b"250 OK")
 
-        self.send("250-stream-status=")
-        self.send("250 OK")
+        self.send(b"250-stream-status=")
+        self.send(b"250 OK")
 
-        self.send("250+address-mappings/all=")
-        self.send('.')
-        self.send('250 OK')
+        self.send(b"250+address-mappings/all=")
+        self.send(b".")
+        self.send(b"250 OK")
 
         for ignored in self.state.event_map.items():
-            self.send("250 OK")
+            self.send(b"250 OK")
 
-        self.send("250-entry-guards=")
-        self.send("250 OK")
+        self.send(b"250-entry-guards=")
+        self.send(b"250 OK")
 
-        self.send("250 OK")
+        self.send(b"250 OK")
 
         self.assertTrue(self.state.find_circuit(123))
-        self.assertEquals(len(self.state.circuits), 1)
+        self.assertEqual(len(self.state.circuits), 1)
 
         return d
 
     def test_unset_attacher(self):
 
+        @implementer(IStreamAttacher)
         class MyAttacher(object):
-            implements(IStreamAttacher)
 
             def attach_stream(self, stream, circuits):
                 return None
 
         fr = FakeReactor(self)
-        self.state.set_attacher(MyAttacher(), fr)
-        self.send("250 OK")
+        attacher = MyAttacher()
+        self.state.set_attacher(attacher, fr)
+        self.send(b"250 OK")
         self.state.set_attacher(None, fr)
-        self.send("250 OK")
+        self.send(b"250 OK")
         self.assertEqual(
             self.transport.value(),
-            'SETCONF __LeaveStreamsUnattached=1\r\nSETCONF'
-            ' __LeaveStreamsUnattached=0\r\n'
+            b'SETCONF __LeaveStreamsUnattached=1\r\nSETCONF'
+            b' __LeaveStreamsUnattached=0\r\n'
+        )
+
+    def test_attacher_twice(self):
+        """
+        It should be an error to set an attacher twice
+        """
+        @implementer(IStreamAttacher)
+        class MyAttacher(object):
+            pass
+
+        attacher = MyAttacher()
+        self.state.set_attacher(attacher, FakeReactor(self))
+        # attach the *same* instance twice; not an error
+        self.state.set_attacher(attacher, FakeReactor(self))
+        with self.assertRaises(RuntimeError) as ctx:
+            self.state.set_attacher(MyAttacher(), FakeReactor(self))
+        self.assertTrue(
+            "already have an attacher" in str(ctx.exception)
+        )
+
+    @defer.inlineCallbacks
+    def _test_attacher_both_apis(self):
+        """
+        similar to above, but first set_attacher is implicit via
+        Circuit.stream_via
+        """
+        reactor = Mock()
+        directlyProvides(reactor, IReactorCore)
+
+        @implementer(IStreamAttacher)
+        class MyAttacher(object):
+            pass
+
+        circ = Circuit(self.state)
+        circ.state = 'BUILT'
+
+        # use the "preferred" API, which will set an attacher
+        factory = Mock()
+        proto = Mock()
+        proto.when_done = Mock(return_value=defer.succeed(None))
+        factory.connect = Mock(return_value=defer.succeed(proto))
+        ep = circ.stream_via(reactor, 'meejah.ca', 443, factory)
+        addr = Mock()
+        addr.host = '10.0.0.1'
+        addr.port = 1011
+        ep._target_endpoint._get_address = Mock(return_value=defer.succeed(addr))
+        print("EP", ep)
+        attacher = yield _get_circuit_attacher(reactor, self.state)
+        print("attacher", attacher)
+        d = ep.connect('foo')
+        print("doin' it")
+        stream = Mock()
+        import ipaddress
+        stream.source_addr = ipaddress.IPv4Address(u'10.0.0.1')
+        stream.source_port = 1011
+        attacher.attach_stream(stream, [])
+        yield d
+
+        # ...now use the low-level API (should be an error)
+        with self.assertRaises(RuntimeError) as ctx:
+            self.state.set_attacher(MyAttacher(), FakeReactor(self))
+        self.assertTrue(
+            "already have an attacher" in str(ctx.exception)
         )
 
     def test_attacher(self):
+        @implementer(IStreamAttacher)
         class MyAttacher(object):
-            implements(IStreamAttacher)
 
             def __init__(self):
                 self.streams = []
@@ -528,41 +643,41 @@ class StateTests(unittest.TestCase):
         self.protocol._set_valid_events(events)
         self.state._add_events()
         for ignored in self.state.event_map.items():
-            self.send("250 OK")
+            self.send(b"250 OK")
 
-        self.send("650 STREAM 1 NEW 0 ca.yahoo.com:80 SOURCE_ADDR=127.0.0.1:54327 PURPOSE=USER")
-        self.send("650 STREAM 1 REMAP 0 87.248.112.181:80 SOURCE=CACHE")
+        self.send(b"650 STREAM 1 NEW 0 ca.yahoo.com:80 SOURCE_ADDR=127.0.0.1:54327 PURPOSE=USER")
+        self.send(b"650 STREAM 1 REMAP 0 87.248.112.181:80 SOURCE=CACHE")
         self.assertEqual(len(attacher.streams), 1)
         self.assertEqual(attacher.streams[0].id, 1)
         self.assertEqual(len(self.protocol.commands), 1)
-        self.assertEqual(self.protocol.commands[0][1], 'ATTACHSTREAM 1 0')
+        self.assertEqual(self.protocol.commands[0][1], b'ATTACHSTREAM 1 0')
 
         # we should totally ignore .exit URIs
         attacher.streams = []
-        self.send("650 STREAM 2 NEW 0 10.0.0.0.$E11D2B2269CC25E67CA6C9FB5843497539A74FD0.exit:80 SOURCE_ADDR=127.0.0.1:12345 PURPOSE=TIME")
+        self.send(b"650 STREAM 2 NEW 0 10.0.0.0.$E11D2B2269CC25E67CA6C9FB5843497539A74FD0.exit:80 SOURCE_ADDR=127.0.0.1:12345 PURPOSE=TIME")
         self.assertEqual(len(attacher.streams), 0)
         self.assertEqual(len(self.protocol.commands), 1)
 
         # we should NOT ignore .onion URIs
         attacher.streams = []
-        self.send("650 STREAM 3 NEW 0 xxxxxxxxxxxxxxxx.onion:80 SOURCE_ADDR=127.0.0.1:12345 PURPOSE=TIME")
+        self.send(b"650 STREAM 3 NEW 0 xxxxxxxxxxxxxxxx.onion:80 SOURCE_ADDR=127.0.0.1:12345 PURPOSE=TIME")
         self.assertEqual(len(attacher.streams), 1)
         self.assertEqual(len(self.protocol.commands), 2)
-        self.assertEqual(self.protocol.commands[1][1], 'ATTACHSTREAM 3 0')
+        self.assertEqual(self.protocol.commands[1][1], b'ATTACHSTREAM 3 0')
 
         # normal attach
         circ = FakeCircuit(1)
         circ.state = 'BUILT'
         self.state.circuits[1] = circ
         attacher.answer = circ
-        self.send("650 STREAM 4 NEW 0 xxxxxxxxxxxxxxxx.onion:80 SOURCE_ADDR=127.0.0.1:12345 PURPOSE=TIME")
+        self.send(b"650 STREAM 4 NEW 0 xxxxxxxxxxxxxxxx.onion:80 SOURCE_ADDR=127.0.0.1:12345 PURPOSE=TIME")
         self.assertEqual(len(attacher.streams), 2)
         self.assertEqual(len(self.protocol.commands), 3)
-        self.assertEqual(self.protocol.commands[2][1], 'ATTACHSTREAM 4 1')
+        self.assertEqual(self.protocol.commands[2][1], b'ATTACHSTREAM 4 1')
 
     def test_attacher_defer(self):
+        @implementer(IStreamAttacher)
         class MyAttacher(object):
-            implements(IStreamAttacher)
 
             def __init__(self, answer):
                 self.streams = []
@@ -583,19 +698,19 @@ class StateTests(unittest.TestCase):
         self.protocol._set_valid_events(events)
         self.state._add_events()
         for ignored in self.state.event_map.items():
-            self.send("250 OK")
+            self.send(b"250 OK")
 
-        self.send("650 STREAM 1 NEW 0 ca.yahoo.com:80 SOURCE_ADDR=127.0.0.1:54327 PURPOSE=USER")
-        self.send("650 STREAM 1 REMAP 0 87.248.112.181:80 SOURCE=CACHE")
+        self.send(b"650 STREAM 1 NEW 0 ca.yahoo.com:80 SOURCE_ADDR=127.0.0.1:54327 PURPOSE=USER")
+        self.send(b"650 STREAM 1 REMAP 0 87.248.112.181:80 SOURCE=CACHE")
         self.assertEqual(len(attacher.streams), 1)
         self.assertEqual(attacher.streams[0].id, 1)
         self.assertEqual(len(self.protocol.commands), 1)
-        self.assertEqual(self.protocol.commands[0][1], 'ATTACHSTREAM 1 1')
+        self.assertEqual(self.protocol.commands[0][1], b'ATTACHSTREAM 1 1')
 
     @defer.inlineCallbacks
     def test_attacher_errors(self):
+        @implementer(IStreamAttacher)
         class MyAttacher(object):
-            implements(IStreamAttacher)
 
             def __init__(self, answer):
                 self.streams = []
@@ -613,7 +728,7 @@ class StateTests(unittest.TestCase):
         msg = ''
         try:
             yield self.state._maybe_attach(stream)
-        except Exception, e:
+        except Exception as e:
             msg = str(e)
         self.assertTrue('circuit unknown' in msg)
 
@@ -621,7 +736,7 @@ class StateTests(unittest.TestCase):
         msg = ''
         try:
             yield self.state._maybe_attach(stream)
-        except Exception, e:
+        except Exception as e:
             msg = str(e)
         self.assertTrue('only attach to BUILT' in msg)
 
@@ -629,13 +744,13 @@ class StateTests(unittest.TestCase):
         msg = ''
         try:
             yield self.state._maybe_attach(stream)
-        except Exception, e:
+        except Exception as e:
             msg = str(e)
         self.assertTrue('Circuit instance' in msg)
 
     def test_attacher_no_attach(self):
+        @implementer(IStreamAttacher)
         class MyAttacher(object):
-            implements(IStreamAttacher)
 
             def __init__(self):
                 self.streams = []
@@ -650,15 +765,14 @@ class StateTests(unittest.TestCase):
         self.protocol._set_valid_events(events)
         self.state._add_events()
         for ignored in self.state.event_map.items():
-            self.send("250 OK")
+            self.send(b"250 OK")
 
         self.transport.clear()
-        self.send("650 STREAM 1 NEW 0 ca.yahoo.com:80 SOURCE_ADDR=127.0.0.1:54327 PURPOSE=USER")
-        self.send("650 STREAM 1 REMAP 0 87.248.112.181:80 SOURCE=CACHE")
+        self.send(b"650 STREAM 1 NEW 0 ca.yahoo.com:80 SOURCE_ADDR=127.0.0.1:54327 PURPOSE=USER")
+        self.send(b"650 STREAM 1 REMAP 0 87.248.112.181:80 SOURCE=CACHE")
         self.assertEqual(len(attacher.streams), 1)
         self.assertEqual(attacher.streams[0].id, 1)
-        print self.transport.value()
-        self.assertEqual(self.transport.value(), '')
+        self.assertEqual(self.transport.value(), b'')
 
     def test_close_stream_with_id(self):
         stream = Stream(self.state)
@@ -666,7 +780,7 @@ class StateTests(unittest.TestCase):
 
         self.state.streams[1] = stream
         self.state.close_stream(stream)
-        self.assertEqual(self.transport.value(), 'CLOSESTREAM 1 1\r\n')
+        self.assertEqual(self.transport.value(), b'CLOSESTREAM 1 1\r\n')
 
     def test_close_stream_with_stream(self):
         stream = Stream(self.state)
@@ -674,7 +788,7 @@ class StateTests(unittest.TestCase):
 
         self.state.streams[1] = stream
         self.state.close_stream(stream.id)
-        self.assertEqual(self.transport.value(), 'CLOSESTREAM 1 1\r\n')
+        self.assertEqual(self.transport.value(), b'CLOSESTREAM 1 1\r\n')
 
     def test_close_stream_invalid_reason(self):
         stream = Stream(self.state)
@@ -693,7 +807,7 @@ class StateTests(unittest.TestCase):
 
         self.state.circuits[1] = circuit
         self.state.close_circuit(circuit.id)
-        self.assertEqual(self.transport.value(), 'CLOSECIRCUIT 1\r\n')
+        self.assertEqual(self.transport.value(), b'CLOSECIRCUIT 1\r\n')
 
     def test_close_circuit_with_circuit(self):
         circuit = Circuit(self.state)
@@ -701,7 +815,7 @@ class StateTests(unittest.TestCase):
 
         self.state.circuits[1] = circuit
         self.state.close_circuit(circuit)
-        self.assertEqual(self.transport.value(), 'CLOSECIRCUIT 1\r\n')
+        self.assertEqual(self.transport.value(), b'CLOSECIRCUIT 1\r\n')
 
     def test_close_circuit_with_flags(self):
         circuit = Circuit(self.state)
@@ -714,7 +828,7 @@ class StateTests(unittest.TestCase):
 
         self.state.circuits[1] = circuit
         self.state.close_circuit(circuit.id, IfUnused=True)
-        self.assertEqual(self.transport.value(), 'CLOSECIRCUIT 1 IfUnused\r\n')
+        self.assertEqual(self.transport.value(), b'CLOSECIRCUIT 1 IfUnused\r\n')
 
     def test_circuit_destroy(self):
         self.state._circuit_update('365 LAUNCHED PURPOSE=GENERAL')
@@ -735,7 +849,7 @@ class StateTests(unittest.TestCase):
         self.protocol._set_valid_events(events)
         self.state._add_events()
         for ignored in self.state.event_map.items():
-            self.send("250 OK")
+            self.send(b"250 OK")
 
         # we use this router later on in an EXTEND
         self.state._update_network_status("""ns/all=
@@ -750,22 +864,23 @@ p reject 1-65535""")
                     ]
         listen = CircuitListener(expected)
         # first add a Circuit before we listen
-        self.protocol.dataReceived("650 CIRC 123 LAUNCHED PURPOSE=GENERAL\r\n")
+        self.protocol.dataReceived(b"650 CIRC 123 LAUNCHED PURPOSE=GENERAL\r\n")
         self.assertEqual(len(self.state.circuits), 1)
 
         # make sure we get added to existing circuits
         self.state.add_circuit_listener(listen)
-        self.assertTrue(listen in self.state.circuits.values()[0].listeners)
+        first_circuit = list(self.state.circuits.values())[0]
+        self.assertTrue(listen in first_circuit.listeners)
 
         # now add a Circuit after we started listening
-        self.protocol.dataReceived("650 CIRC 456 LAUNCHED PURPOSE=GENERAL\r\n")
+        self.protocol.dataReceived(b"650 CIRC 456 LAUNCHED PURPOSE=GENERAL\r\n")
         self.assertEqual(len(self.state.circuits), 2)
-        self.assertTrue(listen in self.state.circuits.values()[0].listeners)
-        self.assertTrue(listen in self.state.circuits.values()[1].listeners)
+        self.assertTrue(listen in list(self.state.circuits.values())[0].listeners)
+        self.assertTrue(listen in list(self.state.circuits.values())[1].listeners)
 
         # now update the first Circuit to ensure we're really, really
         # listening
-        self.protocol.dataReceived("650 CIRC 123 EXTENDED $D82183B1C09E1D7795FF2D7116BAB5106AA3E60E~PPrivCom012 PURPOSE=GENERAL\r\n")
+        self.protocol.dataReceived(b"650 CIRC 123 EXTENDED $D82183B1C09E1D7795FF2D7116BAB5106AA3E60E~PPrivCom012 PURPOSE=GENERAL\r\n")
         self.assertEqual(len(listen.expected), 0)
 
     def test_router_from_id_invalid_key(self):
@@ -819,8 +934,10 @@ p accept 43,53,79-81,110,143,194,220,443,953,989-990,993,995,1194,1293,1723,1863
 .''')
             self.fail()
 
-        except RuntimeError, e:
-            self.assertTrue('"s "' in str(e))
+        except RuntimeError as e:
+            # self.assertTrue('Illegal state' in str(e))
+            # flip back when we go back to Automat
+            self.assertTrue('Expected "s "' in str(e))
 
     def test_routers_no_policy(self):
         """
@@ -899,48 +1016,49 @@ p accept 43,53,79-81,110,143,194,220,443,953,989-990,993,995,1194,1293,1723,1863
         self.protocol._set_valid_events(' '.join(self.state.event_map.keys()))
         self.state._bootstrap()
 
-        self.send("250+ns/all=")
-        self.send(".")
-        self.send("250 OK")
+        self.send(b"250+ns/all=")
+        self.send(b".")
+        self.send(b"250 OK")
 
-        self.send("250+circuit-status=")
-        self.send(".")
-        self.send("250 OK")
+        self.send(b"250+circuit-status=")
+        self.send(b".")
+        self.send(b"250 OK")
 
-        self.send("250-stream-status=")
-        self.send("250 OK")
+        self.send(b"250-stream-status=")
+        self.send(b"250 OK")
 
-        self.send("250-address-mappings/all=")
-        self.send('250 OK')
+        self.send(b"250-address-mappings/all=")
+        self.send(b'250 OK')
 
         for ignored in self.state.event_map.items():
-            self.send("250 OK")
+            self.send(b"250 OK")
 
-        self.send("250-entry-guards=")
-        self.send("250 OK")
+        self.send(b"250-entry-guards=")
+        self.send(b"250 OK")
 
-        self.send("250 OK")
+        self.send(b"250 OK")
 
         # state is now bootstrapped, we can send our NEWCONSENSUS update
 
-        self.protocol.dataReceived('\r\n'.join('''650+NEWCONSENSUS
+        self.protocol.dataReceived(b'\r\n'.join(b'''650+NEWCONSENSUS
 r Unnamed ABJlguUFz1lvQS0jq8nhTdRiXEk /zIVUg1tKMUeyUBoyimzorbQN9E 2012-05-23 01:10:22 219.94.255.254 9001 0
 s Fast Guard Running Stable Valid
 w Bandwidth=166
 p reject 1-65535
 .
 650 OK
-'''.split('\n')))
+'''.split(b'\n')))
 
-        self.protocol.dataReceived('\r\n'.join('''650+NEWCONSENSUS
+        self.protocol.dataReceived(b'\r\n'.join(b'''650+NEWCONSENSUS
 r Unnamed ABJlguUFz1lvQS0jq8nhTdRiXEk /zIVUg1tKMUeyUBoyimzorbQN9E 2012-05-23 01:10:22 219.94.255.254 9001 0
 s Fast Guard Running Stable Valid
 w Bandwidth=166
 p reject 1-65535
 .
 650 OK
-'''.split('\n')))
+'''.split(b'\n')))
 
+        self.assertEqual(1, len(self.state.all_routers))
         self.assertTrue('Unnamed' in self.state.routers)
         self.assertTrue('$00126582E505CF596F412D23ABC9E14DD4625C49' in self.state.routers)
 
@@ -956,37 +1074,37 @@ p reject 1-65535
         self.protocol._set_valid_events(' '.join(self.state.event_map.keys()))
         self.state._bootstrap()
 
-        self.send("250+ns/all=")
-        self.send(".")
-        self.send("250 OK")
+        self.send(b"250+ns/all=")
+        self.send(b".")
+        self.send(b"250 OK")
 
-        self.send("250+circuit-status=")
-        self.send(".")
-        self.send("250 OK")
+        self.send(b"250+circuit-status=")
+        self.send(b".")
+        self.send(b"250 OK")
 
-        self.send("250-stream-status=")
-        self.send("250 OK")
+        self.send(b"250-stream-status=")
+        self.send(b"250 OK")
 
-        self.send("250-address-mappings/all=")
-        self.send('250 OK')
+        self.send(b"250-address-mappings/all=")
+        self.send(b"250 OK")
 
         for ignored in self.state.event_map.items():
-            self.send("250 OK")
+            self.send(b"250 OK")
 
-        self.send("250-entry-guards=")
-        self.send("250 OK")
+        self.send(b"250-entry-guards=")
+        self.send(b"250 OK")
 
-        self.send("250 OK")
+        self.send(b"250 OK")
 
         # state is now bootstrapped, we can send our NEWCONSENSUS update
 
-        self.protocol.dataReceived('\r\n'.join('''650+NEWCONSENSUS
+        self.protocol.dataReceived(b'\r\n'.join(b'''650+NEWCONSENSUS
 r Unnamed ABJlguUFz1lvQS0jq8nhTdRiXEk /zIVUg1tKMUeyUBoyimzorbQN9E 2012-05-23 01:10:22 219.94.255.254 9001 0
 s Fast Guard Running Stable Valid
 w Bandwidth=166
 .
 650 OK
-'''.split('\n')))
+'''.split(b'\n')))
 
         self.assertTrue('Unnamed' in self.state.routers)
         self.assertTrue('$00126582E505CF596F412D23ABC9E14DD4625C49' in self.state.routers)
@@ -1003,36 +1121,36 @@ w Bandwidth=166
         self.protocol._set_valid_events(' '.join(self.state.event_map.keys()))
         self.state._bootstrap()
 
-        self.send("250+ns/all=")
-        self.send(".")
-        self.send("250 OK")
+        self.send(b"250+ns/all=")
+        self.send(b".")
+        self.send(b"250 OK")
 
-        self.send("250+circuit-status=")
-        self.send(".")
-        self.send("250 OK")
+        self.send(b"250+circuit-status=")
+        self.send(b".")
+        self.send(b"250 OK")
 
-        self.send("250-stream-status=")
-        self.send("250 OK")
+        self.send(b"250-stream-status=")
+        self.send(b"250 OK")
 
-        self.send("250-address-mappings/all=")
-        self.send('250 OK')
+        self.send(b"250-address-mappings/all=")
+        self.send(b"250 OK")
 
         for ignored in self.state.event_map.items():
-            self.send("250 OK")
+            self.send(b"250 OK")
 
-        self.send("250-entry-guards=")
-        self.send("250 OK")
+        self.send(b"250-entry-guards=")
+        self.send(b"250 OK")
 
-        self.send("250 OK")
+        self.send(b"250 OK")
 
         # state is now bootstrapped, we can send our NEWCONSENSUS update
 
-        self.protocol.dataReceived('\r\n'.join('''650+NEWCONSENSUS
+        self.protocol.dataReceived(b'\r\n'.join(b'''650+NEWCONSENSUS
 r Unnamed ABJlguUFz1lvQS0jq8nhTdRiXEk /zIVUg1tKMUeyUBoyimzorbQN9E 2012-05-23 01:10:22 219.94.255.254 9001 0
 s Fast Guard Running Stable Valid
 .
 650 OK
-'''.split('\n')))
+'''.split(b'\n')))
 
         self.assertTrue('Unnamed' in self.state.routers)
         self.assertTrue('$00126582E505CF596F412D23ABC9E14DD4625C49' in self.state.routers)
@@ -1064,19 +1182,20 @@ s Fast Guard Running Stable Valid
         self.protocol._set_valid_events('CIRC STREAM ORCONN BW DEBUG INFO NOTICE WARN ERR NEWDESC ADDRMAP AUTHDIR_NEWDESCS DESCCHANGED NS STATUS_GENERAL STATUS_CLIENT STATUS_SERVER GUARD STREAM_BW CLIENTS_SEEN NEWCONSENSUS BUILDTIMEOUT_SET')
         self.state._add_events()
         for ignored in self.state.event_map.items():
-            self.send("250 OK")
+            self.send(b"250 OK")
 
         expected = [('new', {}),
                     ]
         listen = StreamListener(expected)
-        self.send("650 STREAM 77 NEW 0 www.yahoo.cn:80 SOURCE_ADDR=127.0.0.1:54315 PURPOSE=USER")
+        self.send(b"650 STREAM 77 NEW 0 www.yahoo.cn:80 SOURCE_ADDR=127.0.0.1:54315 PURPOSE=USER")
         self.state.add_stream_listener(listen)
 
-        self.assertTrue(listen in self.state.streams.values()[0].listeners)
+        self.assertEqual(1, len(self.state.streams.values()))
+        self.assertTrue(listen in list(self.state.streams.values())[0].listeners)
         self.assertEqual(len(self.state.streams), 1)
         self.assertEqual(len(listen.expected), 1)
 
-        self.send("650 STREAM 78 NEW 0 www.yahoo.cn:80 SOURCE_ADDR=127.0.0.1:54315 PURPOSE=USER")
+        self.send(b"650 STREAM 78 NEW 0 www.yahoo.cn:80 SOURCE_ADDR=127.0.0.1:54315 PURPOSE=USER")
         self.assertEqual(len(self.state.streams), 2)
         self.assertEqual(len(listen.expected), 0)
 
@@ -1094,18 +1213,18 @@ s Fast Guard Running Stable Valid
         path[0].flags = ['guard']
 
         self.state.build_circuit(path, using_guards=True)
-        self.assertEqual(self.transport.value(), 'EXTENDCIRCUIT 0 0000000000000000000000000000000000000000,0000000000000000000000000000000000000001,0000000000000000000000000000000000000002\r\n')
+        self.assertEqual(self.transport.value(), b'EXTENDCIRCUIT 0 0000000000000000000000000000000000000000,0000000000000000000000000000000000000001,0000000000000000000000000000000000000002\r\n')
         # should have gotten a warning about this not being an entry
         # guard
         self.assertEqual(len(self.flushWarnings()), 1)
 
     def test_build_circuit_no_routers(self):
         self.state.build_circuit()
-        self.assertEqual(self.transport.value(), 'EXTENDCIRCUIT 0\r\n')
+        self.assertEqual(self.transport.value(), b'EXTENDCIRCUIT 0\r\n')
 
     def test_build_circuit_unfound_router(self):
-        self.state.build_circuit(routers=['AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'], using_guards=False)
-        self.assertEqual(self.transport.value(), 'EXTENDCIRCUIT 0 AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\r\n')
+        self.state.build_circuit(routers=[b'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'], using_guards=False)
+        self.assertEqual(self.transport.value(), b'EXTENDCIRCUIT 0 AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\r\n')
 
     def circuit_callback(self, circ):
         self.assertTrue(isinstance(circ, Circuit))
@@ -1129,8 +1248,8 @@ s Fast Guard Running Stable Valid
 
         d = self.state.build_circuit(path, using_guards=True)
         d.addCallback(self.circuit_callback)
-        self.assertEqual(self.transport.value(), 'EXTENDCIRCUIT 0 0000000000000000000000000000000000000000,0000000000000000000000000000000000000001,0000000000000000000000000000000000000002\r\n')
-        self.send('250 EXTENDED 1234')
+        self.assertEqual(self.transport.value(), b'EXTENDCIRCUIT 0 0000000000000000000000000000000000000000,0000000000000000000000000000000000000001,0000000000000000000000000000000000000002\r\n')
+        self.send(b"250 EXTENDED 1234")
         # should have gotten a warning about this not being an entry
         # guard
         self.assertEqual(len(self.flushWarnings()), 1)
@@ -1144,7 +1263,7 @@ s Fast Guard Running Stable Valid
         try:
             self.state._find_circuit_after_extend("FOO 1234")
             self.assertTrue(False)
-        except RuntimeError, e:
+        except RuntimeError as e:
             self.assertTrue('Expected EXTENDED' in str(e))
 
     def test_listener_mixins(self):
@@ -1229,9 +1348,9 @@ s Fast Guard Running Stable Valid
         d = build_timeout_circuit(self.state, clock, path, timeout, using_guards=True)
         d.addCallback(self.circuit_callback)
 
-        self.assertEqual(self.transport.value(), 'EXTENDCIRCUIT 0 0000000000000000000000000000000000000000,0000000000000000000000000000000000000001,0000000000000000000000000000000000000002\r\n')
-        self.send('250 EXTENDED 1234')
-        # we can't just .send('650 CIRC 1234 BUILT') this because we
+        self.assertEqual(self.transport.value(), b'EXTENDCIRCUIT 0 0000000000000000000000000000000000000000,0000000000000000000000000000000000000001,0000000000000000000000000000000000000002\r\n')
+        self.send(b"250 EXTENDED 1234")
+        # we can't just .send(b'650 CIRC 1234 BUILT') this because we
         # didn't fully hook up the protocol to the state, e.g. via
         # post_bootstrap etc.
         self.state.circuits[1234].update(['1234', 'BUILT'])
