@@ -1,9 +1,15 @@
+import os
+import sys
+import tempfile
 from mock import patch
+from unittest import skipIf
+import ipaddress
+
 from twisted.trial import unittest
 from twisted.internet import defer
 from twisted.internet.endpoints import TCP4ServerEndpoint
 from twisted.internet.interfaces import IProtocolFactory
-from zope.interface import implements
+from zope.interface import implementer
 
 from txtorcon.util import process_from_address
 from txtorcon.util import delete_file_or_tree
@@ -12,17 +18,19 @@ from txtorcon.util import ip_from_int
 from txtorcon.util import find_tor_binary
 from txtorcon.util import maybe_ip_addr
 from txtorcon.util import unescape_quoted_string
-
-import os
-import tempfile
+from txtorcon.util import available_tcp_port
+from txtorcon.util import version_at_least
+from txtorcon.util import default_control_port
+from txtorcon.util import _Listener, _ListenerCollection
+from txtorcon.util import create_tbb_web_headers
 
 
 class FakeState:
     tor_pid = 0
 
 
+@implementer(IProtocolFactory)
 class FakeProtocolFactory:
-    implements(IProtocolFactory)
 
     def doStart(self):
         "IProtocolFactory API"
@@ -57,14 +65,15 @@ class TestGeoIpDatabaseLoading(unittest.TestCase):
         ret_val = util.create_geoip(f)
         delete_file_or_tree(f)
         util.GeoIP = _GeoIP
-        self.assertEquals(ret_val, None)
+        self.assertEqual(ret_val, None)
 
+    @skipIf('pypy' in sys.version.lower(), "No GeoIP in PyPy")
     def test_return_geoip_object(self):
         from txtorcon import util
         (fd, f) = tempfile.mkstemp()
         ret_val = util.create_geoip(f)
         delete_file_or_tree(f)
-        self.assertEquals(type(ret_val).__name__, 'GeoIP')
+        self.assertEqual(type(ret_val).__name__, 'GeoIP')
 
 
 class TestFindKeywords(unittest.TestCase):
@@ -102,8 +111,8 @@ class TestNetLocation(unittest.TestCase):
             util.city = FakeGeoIP(version=2)
             nl = util.NetLocation('127.0.0.1')
             self.assertTrue(nl.city)
-            self.assertEquals(nl.city[0], 'City')
-            self.assertEquals(nl.city[1], 'Region')
+            self.assertEqual(nl.city[0], 'City')
+            self.assertEqual(nl.city[1], 'Region')
         finally:
             util.ity = orig
 
@@ -114,8 +123,8 @@ class TestNetLocation(unittest.TestCase):
             util.city = FakeGeoIP(version=3)
             nl = util.NetLocation('127.0.0.1')
             self.assertTrue(nl.city)
-            self.assertEquals(nl.city[0], 'City')
-            self.assertEquals(nl.city[1], 'Region')
+            self.assertEqual(nl.city[0], 'City')
+            self.assertEqual(nl.city[1], 'Region')
         finally:
             util.ity = orig
 
@@ -208,11 +217,12 @@ class TestProcessFromUtil(unittest.TestCase):
         # tests, but I was using "nc" before, and I think this is
         # preferable.
         from twisted.internet import reactor
-        ep = TCP4ServerEndpoint(reactor, 9887)
+        port = yield available_tcp_port(reactor)
+        ep = TCP4ServerEndpoint(reactor, port)
         listener = yield ep.listen(FakeProtocolFactory())
 
         try:
-            pid = process_from_address('0.0.0.0', 9887, self.fakestate)
+            pid = process_from_address('0.0.0.0', port, self.fakestate)
         finally:
             listener.stopListening()
 
@@ -223,7 +233,7 @@ class TestDelete(unittest.TestCase):
 
     def test_delete_file(self):
         (fd, f) = tempfile.mkstemp()
-        os.write(fd, 'some\ndata\n')
+        os.write(fd, b'some\ndata\n')
         os.close(fd)
         self.assertTrue(os.path.exists(f))
         delete_file_or_tree(f)
@@ -231,8 +241,8 @@ class TestDelete(unittest.TestCase):
 
     def test_delete_tree(self):
         d = tempfile.mkdtemp()
-        f = open(os.path.join(d, 'foo'), 'w')
-        f.write('foo\n')
+        f = open(os.path.join(d, 'foo'), 'wb')
+        f.write(b'foo\n')
         f.close()
 
         self.assertTrue(os.path.exists(d))
@@ -268,16 +278,17 @@ class TestFindTor(unittest.TestCase):
 
 class TestIpAddr(unittest.TestCase):
 
-    @patch('txtorcon.util.ipaddress')
-    def test_create_ipaddr(self, ipaddr):
+    def test_create_ipaddr(self):
         ip = maybe_ip_addr('1.2.3.4')
+        self.assertTrue(isinstance(ip, ipaddress.IPv4Address))
 
     @patch('txtorcon.util.ipaddress')
-    def test_create_ipaddr(self, ipaddr):
+    def test_create_ipaddr_fail(self, ipaddr):
         def foo(blam):
             raise ValueError('testing')
         ipaddr.ip_address.side_effect = foo
         ip = maybe_ip_addr('1.2.3.4')
+        self.assertTrue(isinstance(ip, type('1.2.3.4')))
 
 
 class TestUnescapeQuotedString(unittest.TestCase):
@@ -315,17 +326,14 @@ class TestUnescapeQuotedString(unittest.TestCase):
         Octal numbers can be escaped by a backslash:
         \0 is interpreted as a byte with the value 0
         '''
-        for number in range(1000):
-            escaped = '\\{}'.format(number)
+        for number in range(0x7f):
+            escaped = '\\%o' % number
             result = unescape_quoted_string('"{}"'.format(escaped))
-
-            expected = escaped.decode('string-escape')
-            if expected[0] == '\\' and len(expected) > 1:
-                expected = expected[1:]
+            expected = chr(number)
 
             msg = "Number not decoded correctly: {escaped} -> {result} instead of {expected}"
             msg = msg.format(escaped=escaped, result=repr(result), expected=repr(expected))
-            self.assertEquals(result, expected, msg=msg)
+            self.assertEqual(result, expected, msg=msg)
 
     def test_invalid_string_unescaping(self):
         invalid_escaped = [
@@ -337,3 +345,118 @@ class TestUnescapeQuotedString(unittest.TestCase):
 
         for invalid_string in invalid_escaped:
             self.assertRaises(ValueError, unescape_quoted_string, invalid_string)
+
+
+class TestVersions(unittest.TestCase):
+    def test_version_1(self):
+        self.assertTrue(
+            version_at_least("1.2.3.4", 1, 2, 3, 4)
+        )
+
+    def test_version_2(self):
+        self.assertFalse(
+            version_at_least("1.2.3.4", 1, 2, 3, 5)
+        )
+
+    def test_version_3(self):
+        self.assertTrue(
+            version_at_least("1.2.3.4", 1, 2, 3, 2)
+        )
+
+    def test_version_4(self):
+        self.assertTrue(
+            version_at_least("2.1.1.1", 2, 0, 0, 0)
+        )
+
+
+class TestHeaders(unittest.TestCase):
+
+    def test_simple(self):
+        create_tbb_web_headers()
+
+
+class TestDefaultPort(unittest.TestCase):
+
+    def test_no_env_var(self):
+        p = default_control_port()
+        self.assertEqual(p, 9151)
+
+    @patch('txtorcon.util.os')
+    def test_env_var(self, fake_os):
+        fake_os.environ = dict(TX_CONTROL_PORT=1234)
+        p = default_control_port()
+        self.assertEqual(p, 1234)
+
+
+class TestListeners(unittest.TestCase):
+
+    def test_add_remove(self):
+        listener = _Listener()
+        calls = []
+
+        def cb(*args, **kw):
+            calls.append((args, kw))
+
+        listener.add(cb)
+        listener.notify('foo', 'bar', quux='zing')
+        listener.remove(cb)
+        listener.notify('foo', 'bar', quux='zing')
+
+        self.assertEqual(1, len(calls))
+        self.assertEqual(('foo', 'bar'), calls[0][0])
+        self.assertEqual(dict(quux='zing'), calls[0][1])
+
+    def test_notify_with_exception(self):
+        listener = _Listener()
+        calls = []
+
+        def cb(*args, **kw):
+            calls.append((args, kw))
+
+        def bad_cb(*args, **kw):
+            raise Exception("sadness")
+
+        listener.add(bad_cb)
+        listener.add(cb)
+        listener.notify('foo', 'bar', quux='zing')
+
+        self.assertEqual(1, len(calls))
+        self.assertEqual(('foo', 'bar'), calls[0][0])
+        self.assertEqual(dict(quux='zing'), calls[0][1])
+
+    def test_collection_invalid_event(self):
+        collection = _ListenerCollection(['event0', 'event1'])
+
+        with self.assertRaises(Exception) as ctx:
+            collection('bad', lambda: None)
+        self.assertTrue('Invalid event' in str(ctx.exception))
+
+    def test_collection_invalid_event_notify(self):
+        collection = _ListenerCollection(['event0', 'event1'])
+
+        with self.assertRaises(Exception) as ctx:
+            collection.notify('bad', lambda: None)
+        self.assertTrue('Invalid event' in str(ctx.exception))
+
+    def test_collection_invalid_event_remove(self):
+        collection = _ListenerCollection(['event0', 'event1'])
+
+        with self.assertRaises(Exception) as ctx:
+            collection.remove('bad', lambda: None)
+        self.assertTrue('Invalid event' in str(ctx.exception))
+
+    def test_collection(self):
+        collection = _ListenerCollection(['event0', 'event1'])
+        calls = []
+
+        def cb(*args, **kw):
+            calls.append((args, kw))
+
+        collection('event0', cb)
+        collection.notify('event0', 'foo', 'bar', quux='zing')
+        collection.remove('event0', cb)
+        collection.notify('event0', 'foo', 'bar', quux='zing')
+
+        self.assertEqual(1, len(calls))
+        self.assertEqual(calls[0][0], ('foo', 'bar'))
+        self.assertEqual(calls[0][1], dict(quux='zing'))
