@@ -11,26 +11,10 @@ import tempfile
 import functools
 
 from txtorcon.util import available_tcp_port
+from txtorcon.socks import TorSocksEndpoint
 
-# backwards-compatibility dance: we "should" be using the
-# ...WithReactor class, but in Twisted prior to 14, there is no such
-# class (and the parse() doesn't provide a 'reactor' argument).
-try:
-    from twisted.internet.interfaces import IStreamClientEndpointStringParserWithReactor
-    _HAVE_TX_14 = True
-except ImportError:
-    from twisted.internet.interfaces import IStreamClientEndpointStringParser as IStreamClientEndpointStringParserWithReactor
-    _HAVE_TX_14 = False
-
-try:
-    from twisted.internet.ssl import optionsForClientTLS
-    from txsocksx.tls import TLSWrapClientEndpoint
-    _HAVE_TLS = True
-except ImportError:
-    _HAVE_TLS = False
-
-
-from twisted.internet import defer, reactor
+from twisted.internet.interfaces import IStreamClientEndpointStringParserWithReactor
+from twisted.internet import defer, error
 from twisted.python import log
 from twisted.internet.interfaces import IStreamServerEndpointStringParser
 from twisted.internet.interfaces import IStreamServerEndpoint
@@ -40,17 +24,17 @@ from twisted.internet.interfaces import IAddress
 from twisted.internet.endpoints import serverFromString
 from twisted.internet.endpoints import clientFromString
 from twisted.internet.endpoints import TCP4ClientEndpoint
-from twisted.internet import error
+# from twisted.internet.endpoints import UNIXClientEndpoint
+# from twisted.internet import error
 from twisted.plugin import IPlugin
 from twisted.python.util import FancyEqMixin
 
 from zope.interface import implementer
 from zope.interface import Interface, Attribute
 
-from txsocksx.client import SOCKS5ClientEndpoint
-
 from .torconfig import TorConfig, launch_tor, HiddenService
-from .torstate import build_tor_connection
+from .torconfig import _endpoint_from_socksport_line
+from .util import SingleObserver
 
 
 _global_tor_config = None
@@ -188,7 +172,7 @@ class TCPHiddenServiceEndpoint(object):
 
 
     :ivar hiddenServiceDir: the data directory, either passed in or created
-        with ``tempfile.mkstemp``
+        with ``tempfile.mkdtemp``
 
     """
 
@@ -208,14 +192,11 @@ class TCPHiddenServiceEndpoint(object):
             only have Group access. XXX FIXME re-test
         """
 
-        @defer.inlineCallbacks
-        def _connect():
-            tor_protocol = yield build_tor_connection(control_endpoint,
-                                                      build_state=False)
-            config = TorConfig(tor_protocol)
-            yield config.post_bootstrap
-            defer.returnValue(config)
-        return TCPHiddenServiceEndpoint(reactor, _connect(), public_port,
+        from txtorcon.controller import connect
+        tor = connect(reactor, control_endpoint)
+        tor.addCallback(lambda t: t.get_config())
+        # tor is a Deferred
+        return TCPHiddenServiceEndpoint(reactor, tor, public_port,
                                         hidden_service_dir=hidden_service_dir,
                                         local_port=local_port)
 
@@ -235,7 +216,7 @@ class TCPHiddenServiceEndpoint(object):
         unless you have a specific need to.
 
         You can also access this global txtorcon instance via
-        :method:`txtorcon.get_global_tor` (which is precisely what
+        :meth:`txtorcon.get_global_tor` (which is precisely what
         this method uses to get it).
 
         All keyword options have defaults (e.g. random ports, or
@@ -299,11 +280,7 @@ class TCPHiddenServiceEndpoint(object):
             :api:`twisted.internet.interfaces.IReactorTCP` provider
 
         :param config:
-            :class:`txtorcon.TorConfig` instance (doesn't need to be
-            bootstrapped) or a Deferred. Note that ``save()`` will be
-            called on this at least once. FIXME should I just accept a
-            TorControlProtocol instance instead, and create my own
-            TorConfig?
+            :class:`txtorcon.TorConfig` instance.
 
         :param public_port:
             The port number we will advertise in the hidden serivces
@@ -316,7 +293,7 @@ class TCPHiddenServiceEndpoint(object):
         :param hidden_service_dir:
             If not None, point to a HiddenServiceDir directory
             (i.e. with "hostname" and "private_key" files in it). If
-            not provided, one is created with temp.mkstemp() AND
+            not provided, one is created with temp.mkdtemp() AND
             DELETED when the reactor shuts down.
 
         :param stealth_auth:
@@ -346,7 +323,7 @@ class TCPHiddenServiceEndpoint(object):
             self.hidden_service_dir = tempfile.mkdtemp(prefix='tortmp')
             log.msg('Will delete "%s" at shutdown.' % self.hidden_service_dir)
             delete = functools.partial(shutil.rmtree, self.hidden_service_dir)
-            reactor.addSystemEventTrigger('before', 'shutdown', delete)
+            self.reactor.addSystemEventTrigger('before', 'shutdown', delete)
 
     @property
     def onion_uri(self):
@@ -372,12 +349,16 @@ class TCPHiddenServiceEndpoint(object):
 
     def _tor_progress_update(self, prog, tag, summary):
         log.msg('%d%% %s' % (prog, summary))
+        # we re-adjust the percentage-scale, using 105% and 110% for
+        # the two parts of waiting for descriptor upload. That is, we
+        # want: 110 * constant == 100.0
         for p in self.progress_listeners:
-            p(prog, tag, summary)
+            p(prog * (100.0 / 110.0), tag, summary)
 
     @defer.inlineCallbacks
     def listen(self, protocolfactory):
-        """Implement :api:`twisted.internet.interfaces.IStreamServerEndpoint
+        """
+        Implement :api:`twisted.internet.interfaces.IStreamServerEndpoint
         <IStreamServerEndpoint>`.
 
         Returns a Deferred that delivers an
@@ -597,7 +578,7 @@ class TCPHiddenServiceEndpointParser(object):
     the OS.
 
     If ``hiddenServiceDir`` is not specified, one is created with
-    ``tempfile.mkstemp()``. The IStreamServerEndpoint returned will be
+    ``tempfile.mkdtemp()``. The IStreamServerEndpoint returned will be
     an instance of :class:`txtorcon.TCPHiddenServiceEndpoint`
     """
     prefix = "onion"
@@ -641,49 +622,135 @@ class TCPHiddenServiceEndpointParser(object):
                                                    control_port=controlPort)
 
 
+@defer.inlineCallbacks
+def _create_socks_endpoint(reactor, control_protocol, socks_config=None):
+    """
+    Internal helper.
+
+    This uses an already-configured SOCKS endpoint from the attached
+    Tor, or creates a new TCP one (and configures Tor with it). If
+    socks_config is non-None, it is a SOCKSPort line and will either
+    be used if it already exists or will be created.
+    """
+    socks_ports = yield control_protocol.get_conf('SOCKSPort')
+    if socks_ports:
+        socks_ports = list(socks_ports.values())[0]
+        if not isinstance(socks_ports, list):
+            socks_ports = [socks_ports]
+    else:
+        # return from get_conf was an empty dict; we want a list
+        socks_ports = []
+
+    # everything in the SocksPort list can include "options" after the
+    # initial value. We don't care about those, but do need to strip
+    # them.
+    socks_ports = [port.split()[0] for port in socks_ports]
+
+    # could check platform? but why would you have unix ports on a
+    # platform that doesn't?
+    unix_ports = set([p.startswith('unix:') for p in socks_ports])
+    tcp_ports = set(socks_ports) - unix_ports
+
+    socks_endpoint = None
+    for p in list(unix_ports) + list(tcp_ports):  # prefer unix-ports
+        if socks_config and p != socks_config:
+            continue
+        try:
+            socks_endpoint = _endpoint_from_socksport_line(reactor, p)
+        except Exception as e:
+            log.msg("clientFromString('{}') failed: {}".format(p, e))
+
+    # if we still don't have an endpoint, nothing worked (or there
+    # were no SOCKSPort lines at all) so we add config to tor
+    if socks_endpoint is None:
+        if socks_config is None:
+            # is a unix-socket in /tmp on a supported platform better than
+            # this?
+            port = yield available_tcp_port(reactor)
+            socks_config = str(port)
+        socks_ports.append(socks_config)
+
+        # NOTE! We must set all the ports in one command or we'll
+        # destroy pre-existing config
+        args = []
+        for p in socks_ports:
+            args.append('SOCKSPort')
+            args.append(p)
+        yield control_protocol.set_conf(*args)
+        socks_endpoint = _endpoint_from_socksport_line(reactor, socks_config)
+
+    defer.returnValue(socks_endpoint)
+
+
 @implementer(IStreamClientEndpoint)
 class TorClientEndpoint(object):
     """
-    I am an endpoint class who attempts to establish a SOCKS5
-    connection with the system tor process. If no socks_endpoint is
-    given, I will try TCP4 to localhost on ports 9050 then 9150.
+    An IStreamClientEndpoint which establishes a connection via Tor.
 
-    :param socks_endpoint:
-        An IStreamClientEndpoint that will connect to a SOCKS5
-        port. Tor can speak SOCKS5 over either TCP4 or Unix sockets.
+    You should not instantiate these directly; use
+    ``clientFromString()``, :meth:`txtorcon.Tor.stream_via` or
+    :meth:`txtorcon.Circuit.stream_via`
 
-    :param tls:
-        If True, we will attemp TLS negotiation after the SOCKS forwarding
-        is set up.
+    :param host:
+        The hostname to connect to. This of course can be a Tor Hidden
+        Service onion address.
+
+    :param port: The tcp port or Tor Hidden Service port.
+
+    :param socks_endpoint: An IStreamClientEndpoint pointing at (one
+        of) our Tor's SOCKS ports. These can be instantiated with
+        :meth:`txtorcon.TorConfig.socks_endpoint`.
+
+    :param tls: Can be False or True (to get default Browser-like
+        hostname verification) or the result of calling
+        optionsForClientTLS() yourself. Default is True.
     """
-    # XXX should get these via the control connection, i.e. ask Tor
-    # via GETINFO net/listeners/socks or whatever
+
     socks_ports_to_try = [9050, 9150]
 
-    def __init__(self, host, port,
-                 socks_endpoint=None,
+    @classmethod
+    def from_connection(cls, reactor, control_protocol, host, port,
+                        tls=None,
+                        socks_endpoint=None):
+        if socks_endpoint is None:
+            socks_endpoint = _create_socks_endpoint(reactor, control_protocol)
+        return TorClientEndpoint(
+            host, port,
+            socks_endpoint=socks_endpoint,
+            tls=tls,
+            reactor=reactor,
+        )
+
+    def __init__(self,
+                 host, port,
+                 socks_endpoint=None,  # can be Deferred
+                 tls=False,
+
+                 # XXX our custom SOCKS stuff doesn't support auth (yet?)
                  socks_username=None, socks_password=None,
-                 tls=False, **kw):
+                 reactor=None, **kw):
         if host is None or port is None:
             raise ValueError('host and port must be specified')
 
         self.host = host
         self.port = int(port)
-        self.socks_endpoint = socks_endpoint
-        self.socks_username = socks_username
-        self.socks_password = socks_password
-        self.tls = tls
-
-        if self.tls and not _HAVE_TLS:
-            raise ValueError(
-                "'tls=True' but we don't have TLS support"
-            )
+        self._socks_endpoint = socks_endpoint
+        self._socks_username = socks_username
+        self._socks_password = socks_password
+        self._tls = tls
+        # XXX FIXME we 'should' probably include 'reactor' as the
+        # first arg to this class, but technically that's a
+        # breaking change :(
+        self._reactor = reactor
+        if reactor is None:
+            from twisted.internet import reactor
+            self._reactor = reactor
 
         # backwards-compatibility: you used to specify a TCP SOCKS
-        # endpoint via socks_hostname= and socks_port= kwargs
-        if self.socks_endpoint is None:
+        # endpoint via socks_host= and socks_port= kwargs
+        if self._socks_endpoint is None:
             try:
-                self.socks_endpoint = TCP4ClientEndpoint(
+                self._socks_endpoint = TCP4ClientEndpoint(
                     reactor,
                     kw['socks_hostname'],
                     kw['socks_port'],
@@ -696,42 +763,61 @@ class TorClientEndpoint(object):
         # was None but the user specified the (old)
         # socks_hostname/socks_port (in which case we do NOT want
         # guessing_enabled
-        if self.socks_endpoint is None:
+        if self._socks_endpoint is None:
             self._socks_port_iter = iter(self.socks_ports_to_try)
             self._socks_guessing_enabled = True
         else:
             self._socks_guessing_enabled = False
 
+        # XXX think, do we want to expose these like this? Or some
+        # other way (because they're for stream-isolation, not actual
+        # auth)
+        self._socks_username = socks_username
+        self._socks_password = socks_password
+        self._when_address = SingleObserver()
+
+    def _get_address(self):
+        """
+        internal helper.
+
+        *le sigh*. This is basically just to support
+        TorCircuitEndpoint; see TorSocksEndpoint._get_address(). There
+        shouldn't be any need for "actual users" to need this!
+
+        This returns a Deferred that fires once:
+          - we have an underlying SOCKS5 endpoint
+          - ...and it has received a local connection (and hence the address/port)
+        """
+        return self._when_address.when_fired()
+
     @defer.inlineCallbacks
     def connect(self, protocolfactory):
         last_error = None
-        kwargs = dict()
-        if self.socks_username is not None and self.socks_password is not None:
-            kwargs['methods'] = dict(
-                login=(self.socks_username, self.socks_password),
+        # XXX fix in socks.py stuff for socks_username, socks_password
+        if self._socks_username or self._socks_password:
+            raise RuntimeError(
+                "txtorcon socks support doesn't yet do username/password"
             )
-        if self.socks_endpoint is not None:
-            args = (self.host, self.port, self.socks_endpoint)
-            socks_ep = SOCKS5ClientEndpoint(*args, **kwargs)
-            if self.tls:
-                context = optionsForClientTLS(unicode(self.host))
-                socks_ep = TLSWrapClientEndpoint(context, socks_ep)
+        if self._socks_endpoint is not None:
+            socks_ep = TorSocksEndpoint(
+                self._socks_endpoint,
+                self.host, self.port,
+                self._tls,
+            )
+            # forward the address to any listeners we have
+            socks_ep._get_address().addCallback(self._when_address.fire)
             proto = yield socks_ep.connect(protocolfactory)
             defer.returnValue(proto)
         else:
             for socks_port in self._socks_port_iter:
                 tor_ep = TCP4ClientEndpoint(
-                    reactor,
-                    "127.0.0.1",
+                    self._reactor,
+                    "127.0.0.1",  # XXX socks_hostname, no?
                     socks_port,
                 )
-                args = (self.host, self.port, tor_ep)
-                socks_ep = SOCKS5ClientEndpoint(*args, **kwargs)
-                if self.tls:
-                    # XXX only twisted 14+
-                    context = optionsForClientTLS(unicode(self.host))
-                    socks_ep = TLSWrapClientEndpoint(context, socks_ep)
-
+                socks_ep = TorSocksEndpoint(tor_ep, self.host, self.port, self._tls)
+                # forward the address to any listeners we have
+                socks_ep._get_address().addCallback(self._when_address.fire)
                 try:
                     proto = yield socks_ep.connect(protocolfactory)
                     defer.returnValue(proto)
@@ -763,8 +849,7 @@ class TorClientEndpointStringParser(object):
     ``tor:host=torproject.org:port=443:socksUsername=foo:socksPassword=bar``
 
     If ``socksPort`` is specified, it means only use that port to
-    attempt to proxy through Tor. If unspecified then try some likely
-    socksPorts such as [9050, 9150].
+    attempt to proxy through Tor. If unspecified, we ... XXX?
 
     NOTE that I'm using camelCase variable names in the endpoint
     string to be consistent with the rest of Twisted's naming (and
@@ -776,19 +861,18 @@ class TorClientEndpointStringParser(object):
     """
     prefix = "tor"
 
-    def _parseClient(self, host=None, port=None,
+    def _parseClient(self, reactor,
+                     host=None, port=None,
                      socksHostname=None, socksPort=None,
                      socksUsername=None, socksPassword=None):
         if port is not None:
             port = int(port)
-        if socksHostname is None:
-            socksHostname = '127.0.0.1'
-        if socksPort is not None:
-            socksPort = int(socksPort)
 
         ep = None
         if socksPort is not None:
-            ep = TCP4ClientEndpoint(reactor, socksHostname, socksPort)
+            # Tor can speak SOCKS over unix, too, but this doesn't let
+            # us pass one ...
+            ep = TCP4ClientEndpoint(reactor, socksHostname, int(socksPort))
         return TorClientEndpoint(
             host, port,
             socks_endpoint=ep,
@@ -799,6 +883,4 @@ class TorClientEndpointStringParser(object):
     def parseStreamClient(self, *args, **kwargs):
         # for Twisted 14 and 15 (and more) the first argument is
         # 'reactor', for older Twisteds it's not
-        if _HAVE_TX_14:
-            return self._parseClient(*args[1:], **kwargs)
         return self._parseClient(*args, **kwargs)
