@@ -5,17 +5,26 @@ from __future__ import print_function
 from __future__ import unicode_literals
 from __future__ import with_statement
 
+import json
 from datetime import datetime
 from .util import NetLocation
-from .util import basestring
+import six
+from base64 import b64encode, b64decode
+from binascii import b2a_hex, a2b_hex
+
+from twisted.internet.defer import inlineCallbacks, returnValue
+from twisted.web.client import readBody
 
 
 def hexIdFromHash(thehash):
     """
     From the base-64 encoded hashes Tor uses, this produces the longer
     hex-encoded hashes.
+
+    :param thehash: base64-encoded str
+    :return: hex-encoded hash
     """
-    return "$" + (thehash + "=").decode("base64").encode("hex").upper()
+    return '$' + b2a_hex(b64decode(thehash + '=')).decode('ascii').upper()
 
 
 def hashFromHexId(hexid):
@@ -24,7 +33,7 @@ def hashFromHexId(hexid):
     """
     if hexid[0] == '$':
         hexid = hexid[1:]
-    return hexid.decode("hex").encode("base64")[:-2]
+    return b64encode(a2b_hex(hexid))[:-1].decode('ascii')
 
 
 class PortRange(object):
@@ -35,10 +44,8 @@ class PortRange(object):
         self.min = a
         self.max = b
 
-    def __cmp__(self, b):
-        if b >= self.min and b <= self.max:
-            return 0
-        return 1
+    def __eq__(self, b):
+        return b >= self.min and b <= self.max
 
     def __str__(self):
         return "%d-%d" % (self.min, self.max)
@@ -79,6 +86,14 @@ class Router(object):
 
     @property
     def modified(self):
+        """
+        This is the time of 'the publication time of its most recent
+        descriptor' (in UTC).
+
+        See also dir-spec.txt.
+
+        """
+        # "... in the form YYYY-MM-DD HH:MM:SS, in UTC"
         if self._modified is None:
             self._modified = datetime.strptime(
                 self._modified_unparsed,
@@ -99,6 +114,9 @@ class Router(object):
         self._location = None
 
         self.id_hex = hexIdFromHash(self.id_hash)
+        # for py3, these should be valid (but *not* py2)
+        # assert type(idhash) is not bytes
+        # assert type(orhash) is not bytes
 
     @property
     def location(self):
@@ -139,7 +157,7 @@ class Router(object):
         There is some current work in Twisted for open-ended constants
         (enums) support however, it seems.
         """
-        if isinstance(flags, basestring):
+        if isinstance(flags, (six.text_type, bytes)):
             flags = flags.split()
         self._flags = [x.lower() for x in flags]
         self.name_is_unique = 'named' in self._flags
@@ -153,6 +171,41 @@ class Router(object):
     def bandwidth(self, bw):
         self._bandwidth = int(bw)
 
+    @inlineCallbacks
+    def get_onionoo_details(self, agent):
+        """
+        Requests the 'details' document from onionoo.torproject.org via
+        the given `twisted.web.iweb.IAgent` -- you can get a suitable
+        instance to pass here by calling either :meth:`txtorcon.Tor.web_agent` or
+        :meth:`txtorcon.Circuit.web_agent`.
+        """
+
+        uri = 'https://onionoo.torproject.org/details?lookup={}'.format(self.id_hex[1:]).encode('ascii')
+
+        resp = yield agent.request(b'GET', uri)
+        if resp.code != 200:
+            raise RuntimeError(
+                'Failed to lookup relay details for {}'.format(self.id_hex)
+            )
+        body = yield readBody(resp)
+        data = json.loads(body.decode('ascii'))
+        if len(data['relays']) != 1:
+            raise RuntimeError(
+                'Got multiple relays for {}'.format(self.id_hex)
+            )
+        relay_data = data['relays'][0]
+        if relay_data['fingerprint'].lower() != self.id_hex[1:].lower():
+            raise RuntimeError(
+                'Expected "{}" but got data for "{}"'.format(self.id_hex, relay_data['fingerprint'])
+            )
+        returnValue(relay_data)
+
+    # note that exit-policy is no longer included in the
+    # microdescriptors by default, so this stuff is mostly here as a
+    # historic artifact. If you want to use exit-policy for things
+    # your best bet is to tell your tor to download full descriptors
+    # (SETCONF UseMicrodescriptors 0) instead.
+
     @property
     def policy(self):
         """
