@@ -5,293 +5,47 @@ from __future__ import print_function
 from __future__ import with_statement
 
 import os
-import sys
-import types
+import re
+import six
 import functools
-import tempfile
 import warnings
 from io import StringIO
-import shlex
+from collections import OrderedDict
 
 from twisted.python import log
-from twisted.internet import defer, error, protocol
-from twisted.internet.interfaces import IReactorTime
-from twisted.internet.endpoints import TCP4ClientEndpoint
-from twisted.internet.endpoints import UNIXClientEndpoint
+from twisted.python.compat import nativeString
+from twisted.python.deprecate import deprecated
+from twisted.internet import defer
+from twisted.internet.endpoints import TCP4ClientEndpoint, UNIXClientEndpoint
 
-from txtorcon.torcontrolprotocol import parse_keywords, TorProtocolFactory, DEFAULT_VALUE
-from txtorcon.util import delete_file_or_tree, find_keywords, find_tor_binary
-from txtorcon.log import txtorlog
+from txtorcon.torcontrolprotocol import parse_keywords, DEFAULT_VALUE
+from txtorcon.torcontrolprotocol import TorProtocolError
 from txtorcon.interface import ITorControlProtocol
-if sys.platform in ('linux2', 'darwin'):
-    import pwd
+from txtorcon.util import find_keywords
 
 
-class TorNotFound(RuntimeError):
+class _Version(object):
     """
-    Raised by launch_tor() in case the tor binary was unspecified and could
-    not be found by consulting the shell.
+    Replacement for incremental.Version until
+    https://github.com/meejah/txtorcon/issues/233 and/or
+    https://github.com/hawkowl/incremental/issues/31 is fixed.
     """
+    # as of latest incremental, it should only access .package and
+    # .short() via the getVersionString() method that Twisted's
+    # deprecated() uses...
+
+    def __init__(self, package, major, minor, patch):
+        self.package = package
+        self.major = major
+        self.minor = minor
+        self.patch = patch
+
+    def short(self):
+        return '{}.{}.{}'.format(self.major, self.minor, self.patch)
 
 
-class TorProcessProtocol(protocol.ProcessProtocol):
-
-    def __init__(self, connection_creator, progress_updates=None, config=None,
-                 ireactortime=None, timeout=None, kill_on_stderr=True,
-                 stdout=None, stderr=None):
-        """
-        This will read the output from a Tor process and attempt a
-        connection to its control port when it sees any 'Bootstrapped'
-        message on stdout. You probably don't need to use this
-        directly except as the return value from the
-        :func:`txtorcon.launch_tor` method. tor_protocol contains a
-        valid :class:`txtorcon.TorControlProtocol` instance by that
-        point.
-
-        connection_creator is a callable that should return a Deferred
-        that callbacks with a :class:`txtorcon.TorControlProtocol`;
-        see :func:`txtorcon.launch_tor` for the default one which is a
-        functools.partial that will call
-        ``connect(TorProtocolFactory())`` on an appropriate
-        :api:`twisted.internet.endpoints.TCP4ClientEndpoint`
-
-        :param connection_creator: A no-parameter callable which
-            returns a Deferred which promises a
-            :api:`twisted.internet.interfaces.IStreamClientEndpoint
-            <IStreamClientEndpoint>`. If this is None, we do NOT
-            attempt to connect to the underlying Tor process.
-
-        :param progress_updates: A callback which received progress
-            updates with three args: percent, tag, summary
-
-        :param config: a TorConfig object to connect to the
-            TorControlProtocl from the launched tor (should it succeed)
-
-        :param ireactortime:
-            An object implementing IReactorTime (i.e. a reactor) which
-            needs to be supplied if you pass a timeout.
-
-        :param timeout:
-            An int representing the timeout in seconds. If we are
-            unable to reach 100% by this time we will consider the
-            setting up of Tor to have failed. Must supply ireactortime
-            if you supply this.
-
-        :param kill_on_stderr:
-            When True, kill subprocess if we receive anything on stderr
-
-        :param stdout:
-            Anything subprocess writes to stdout is sent to .write() on this
-
-        :param stderr:
-            Anything subprocess writes to stderr is sent to .write() on this
-
-        :ivar tor_protocol: The TorControlProtocol instance connected
-            to the Tor this :api:`twisted.internet.protocol.ProcessProtocol
-            <ProcessProtocol>`` is speaking to. Will be valid
-            when the `connected_cb` callback runs.
-
-        :ivar connected_cb: Triggered when the Tor process we
-            represent is fully bootstrapped
-
-        """
-
-        self.config = config
-        self.tor_protocol = None
-        self.progress_updates = progress_updates
-
-        if connection_creator:
-            self.connection_creator = connection_creator
-            self.connected_cb = defer.Deferred()
-        else:
-            self.connection_creator = None
-            self.connected_cb = None
-
-        self.attempted_connect = False
-        self.to_delete = []
-        self.kill_on_stderr = kill_on_stderr
-        self.stderr = stderr
-        self.stdout = stdout
-
-        self._setup_complete = False
-        self._did_timeout = False
-        self._timeout_delayed_call = None
-        self._on_exit = []  # Deferred's we owe a call/errback to when we exit
-        if timeout:
-            if not ireactortime:
-                raise RuntimeError(
-                    'Must supply an IReactorTime object when supplying a '
-                    'timeout')
-            ireactortime = IReactorTime(ireactortime)
-            self._timeout_delayed_call = ireactortime.callLater(
-                timeout, self.timeout_expired)
-
-    def quit(self):
-        """
-        This will terminate (with SIGTERM) the underlying Tor process.
-
-        :returns: a Deferred that callback()'s (with None) when the
-            process has actually exited.
-        """
-
-        try:
-            self.transport.signalProcess('TERM')
-            d = defer.Deferred()
-            self._on_exit.append(d)
-
-        except error.ProcessExitedAlready:
-            self.transport.loseConnection()
-            d = defer.succeed(None)
-        return d
-
-    def _signal_on_exit(self, reason):
-        to_notify = self._on_exit
-        self._on_exit = []
-        for d in to_notify:
-            d.callback(None)
-
-    def outReceived(self, data):
-        """
-        :api:`twisted.internet.protocol.ProcessProtocol <ProcessProtocol>` API
-        """
-
-        if self.stdout:
-            self.stdout.write(data)
-
-        # minor hack: we can't try this in connectionMade because
-        # that's when the process first starts up so Tor hasn't
-        # opened any ports properly yet. So, we presume that after
-        # its first output we're good-to-go. If this fails, we'll
-        # reset and try again at the next output (see this class'
-        # tor_connection_failed)
-
-        txtorlog.msg(data)
-        if not self.attempted_connect and self.connection_creator \
-                and 'Bootstrap' in data:
-            self.attempted_connect = True
-            d = self.connection_creator()
-            d.addCallback(self.tor_connected)
-            d.addErrback(self.tor_connection_failed)
-
-    def timeout_expired(self):
-        """
-        A timeout was supplied during setup, and the time has run out.
-        """
-
-        try:
-            self.transport.signalProcess('TERM')
-        except error.ProcessExitedAlready:
-            self.transport.loseConnection()
-        self._did_timeout = True
-
-    def errReceived(self, data):
-        """
-        :api:`twisted.internet.protocol.ProcessProtocol <ProcessProtocol>` API
-        """
-
-        if self.stderr:
-            self.stderr.write(data)
-
-        if self.kill_on_stderr:
-            self.transport.loseConnection()
-            raise RuntimeError(
-                "Received stderr output from slave Tor process: " + data)
-
-    def cleanup(self):
-        """
-        Clean up my temporary files.
-        """
-
-        all([delete_file_or_tree(f) for f in self.to_delete])
-        self.to_delete = []
-
-    def processExited(self, reason):
-        self._signal_on_exit(reason)
-
-    def processEnded(self, status):
-        """
-        :api:`twisted.internet.protocol.ProcessProtocol <ProcessProtocol>` API
-        """
-
-        self.cleanup()
-
-        if status.value.exitCode is None:
-            if self._did_timeout:
-                err = RuntimeError("Timeout waiting for Tor launch..")
-            else:
-                err = RuntimeError(
-                    "Tor was killed (%s)." % status.value.signal)
-        else:
-            err = RuntimeError(
-                "Tor exited with error-code %d" % status.value.exitCode
-            )
-
-        log.err(err)
-        if self.connected_cb:
-            self.connected_cb.errback(err)
-            self.connected_cb = None
-        self._signal_on_exit(status)
-
-    def progress(self, percent, tag, summary):
-        """
-        Can be overridden or monkey-patched if you want to get
-        progress updates yourself.
-        """
-
-        if self.progress_updates:
-            self.progress_updates(percent, tag, summary)
-
-    # the below are all callbacks
-
-    def tor_connection_failed(self, failure):
-        # FIXME more robust error-handling please, like a timeout so
-        # we don't just wait forever after 100% bootstrapped (that
-        # is, we're ignoring these errors, but shouldn't do so after
-        # we'll stop trying)
-        self.attempted_connect = False
-
-    def status_client(self, arg):
-        args = shlex.split(arg)
-        if args[1] != 'BOOTSTRAP':
-            return
-
-        kw = find_keywords(args)
-        prog = int(kw['PROGRESS'])
-        tag = kw['TAG']
-        summary = kw['SUMMARY']
-        self.progress(prog, tag, summary)
-
-        if prog == 100:
-            if self._timeout_delayed_call:
-                self._timeout_delayed_call.cancel()
-                self._timeout_delayed_call = None
-            if self.connected_cb:
-                self.connected_cb.callback(self)
-                self.connected_cb = None
-
-    def tor_connected(self, proto):
-        txtorlog.msg("tor_connected %s" % proto)
-
-        self.tor_protocol = proto
-        if self.config is not None:
-            self.config._update_proto(proto)
-        self.tor_protocol.is_owned = self.transport.pid
-        self.tor_protocol.post_bootstrap.addCallback(
-            self.protocol_bootstrapped).addErrback(
-                self.tor_connection_failed)
-
-    def protocol_bootstrapped(self, proto):
-        txtorlog.msg("Protocol is bootstrapped")
-
-        self.tor_protocol.add_event_listener(
-            'STATUS_CLIENT', self.status_client)
-
-        # FIXME: should really listen for these to complete as well
-        # as bootstrap etc. For now, we'll be optimistic.
-        self.tor_protocol.queue_command('TAKEOWNERSHIP')
-        self.tor_protocol.queue_command('RESETCONF __OwningControllerProcess')
-
-
+@defer.inlineCallbacks
+@deprecated(_Version("txtorcon", 0, 18, 0))
 def launch_tor(config, reactor,
                tor_binary=None,
                progress_updates=None,
@@ -299,219 +53,27 @@ def launch_tor(config, reactor,
                timeout=None,
                kill_on_stderr=True,
                stdout=None, stderr=None):
-    """launches a new Tor process with the given config.
-
-    There may seem to be a ton of options, but don't panic: this
-    method should be easy to use and most options can be ignored
-    except for advanced use-cases. Calling with a completely empty
-    TorConfig should Just Work::
-
-        config = TorConfig()
-        d = launch_tor(config, reactor)
-        d.addCallback(...)
-
-    Note that the incoming TorConfig instance is examined and several
-    config options are acted upon appropriately:
-
-    ``DataDirectory``: if supplied, a tempdir is not created, and the
-    one supplied is not deleted.
-
-    ``ControlPort``: if 0 (zero), a control connection is NOT
-    established (and ``connection_creator`` is ignored). In this case
-    we can't wait for Tor to bootstrap, and **you must kill the tor**
-    yourself.
-
-    ``User``: if this exists, we attempt to set ownership of the tempdir
-    to this user (but only if our effective UID is 0).
-
-    This method may set the following options on the supplied
-    TorConfig object: ``DataDirectory, ControlPort,
-    CookieAuthentication, __OwningControllerProcess`` and WILL call
-    :meth:`txtorcon.TorConfig.save`
-
-    :param config:
-        an instance of :class:`txtorcon.TorConfig` with any
-        configuration values you want.  If ``ControlPort`` isn't set,
-        9052 is used; if ``DataDirectory`` isn't set, tempdir is used
-        to create one (in this case, it will be deleted upon exit).
-
-    :param reactor: a Twisted IReactorCore implementation (usually
-        twisted.internet.reactor)
-
-    :param tor_binary: path to the Tor binary to run. Tries to find the tor
-        binary if unset.
-
-    :param progress_updates: a callback which gets progress updates; gets as
-         args: percent, tag, summary (FIXME make an interface for this).
-
-    :param kill_on_stderr:
-        When True (the default), if Tor prints anything on stderr we
-        kill off the process, close the TorControlProtocol and raise
-        an exception.
-
-    :param stdout: a file-like object to which we write anything that
-        Tor prints on stdout (just needs to support write()).
-
-    :param stderr: a file-like object to which we write anything that
-        Tor prints on stderr (just needs .write()). Note that we kill Tor
-        off by default if anything appears on stderr; pass "no_kill=True"
-        if you don't like the behavior.
-
-    :param connection_creator: is mostly available to ease testing, so
-        you probably don't want to supply this. If supplied, it is a
-        callable that should return a Deferred that delivers an
-        :api:`twisted.internet.interfaces.IProtocol <IProtocol>` or
-        ConnectError.
-        See :api:`twisted.internet.interfaces.IStreamClientEndpoint`.connect
-        Note that this parameter is ignored if config.ControlPort == 0
-
-    :return: a Deferred which callbacks with a TorProcessProtocol
-        connected to the fully-bootstrapped Tor; this has a
-        :class:`txtorcon.TorControlProtocol` instance as `.tor_protocol`. In
-        Tor, ``__OwningControllerProcess`` will be set and TAKEOWNERSHIP will
-        have been called, so if you close the TorControlProtocol the Tor should
-        exit also (see `control-spec
-        <https://gitweb.torproject.org/torspec.git/blob/HEAD:/control-spec.txt>`_
-        3.23). Note that if ControlPort was 0, we don't connect at all
-        and therefore don't wait for Tor to be bootstrapped. In this case, it's
-        up to you to kill off the Tor you created.
-
-    HACKS:
-
-     1. It's hard to know when Tor has both (completely!) written its
-        authentication cookie file AND is listening on the control
-        port. It seems that waiting for the first 'bootstrap' message on
-        stdout is sufficient. Seems fragile...and doesn't work 100% of
-        the time, so FIXME look at Tor source.
     """
+    Deprecated; use launch() instead.
 
-    # We have a slight problem with the approach: we need to pass a
-    # few minimum values to a torrc file so that Tor will start up
-    # enough that we may connect to it. Ideally, we'd be able to
-    # start a Tor up which doesn't really do anything except provide
-    # "AUTHENTICATE" and "GETINFO config/names" so we can do our
-    # config validation.
-
-    # the other option here is to simply write a torrc version of our
-    # config and get Tor to load that...which might be the best
-    # option anyway.
-
-    # actually, can't we pass them all as command-line arguments?
-    # could be pushing some limits for giant configs...
-
-    if tor_binary is None:
-        tor_binary = find_tor_binary()
-    if tor_binary is None:
-        # We fail right here instead of waiting for the reactor to start
-        raise TorNotFound('Tor binary could not be found')
-
-    # make sure we got things that have write() for stderr, stdout
-    # kwargs
-    for arg in [stderr, stdout]:
-        if arg and not getattr(arg, "write", None):
-            raise RuntimeError(
-                'File-like object needed for stdout or stderr args.')
-
-    try:
-        data_directory = config.DataDirectory
-        user_set_data_directory = True
-    except KeyError:
-        user_set_data_directory = False
-        data_directory = tempfile.mkdtemp(prefix='tortmp')
-        config.DataDirectory = data_directory
-
-        # Set ownership on the temp-dir to the user tor will drop privileges to
-        # when executing as root.
-        try:
-            user = config.User
-        except KeyError:
-            pass
-        else:
-            if sys.platform in ('linux2', 'darwin') and os.geteuid() == 0:
-                os.chown(data_directory, pwd.getpwnam(user).pw_uid, -1)
-
-    try:
-        control_port = config.ControlPort
-    except KeyError:
-        control_port = 9052  # FIXME choose a random, unoccupied one?
-        config.ControlPort = control_port
-
-    # so, we support passing in ControlPort=0 -- not really sure if
-    # this is a good idea (since then the caller has to kill the tor
-    # off, etc), but at least one person has requested it :/
-    if control_port != 0:
-        config.CookieAuthentication = 1
-        config.__OwningControllerProcess = os.getpid()
-        if connection_creator is None:
-            if str(control_port).startswith('unix:'):
-                connection_creator = functools.partial(
-                    UNIXClientEndpoint(reactor, control_port[5:]).connect,
-                    TorProtocolFactory()
-                )
-            else:
-                connection_creator = functools.partial(
-                    TCP4ClientEndpoint(reactor, 'localhost', control_port).connect,
-                    TorProtocolFactory()
-                )
-    else:
-        connection_creator = None
-
-    # NOTE well, that if we don't pass "-f" then Tor will merrily load
-    # it's default torrc, and apply our options over top... :/
-    config_args = ['-f', '/non-existant', '--ignore-missing-torrc']
-
-    # ...now add all our config options on the command-line. This
-    # avoids writing a temporary torrc.
-    for (k, v) in config.config_args():
-        config_args.append(k)
-        config_args.append(v)
-
-    # txtorlog.msg('Running with config:\n', ' '.join(config_args))
-
-    process_protocol = TorProcessProtocol(
-        connection_creator,
-        progress_updates,
-        config, reactor,
-        timeout,
-        kill_on_stderr,
-        stdout,
-        stderr
+    See also controller.py
+    """
+    from .controller import launch
+    # XXX FIXME are we dealing with options in the config "properly"
+    # as far as translating semantics from the old launch_tor to
+    # launch()? DataDirectory, User, ControlPort, ...?
+    tor = yield launch(
+        reactor,
+        stdout=stdout,
+        stderr=stderr,
+        progress_updates=progress_updates,
+        tor_binary=tor_binary,
+        connection_creator=connection_creator,
+        timeout=timeout,
+        kill_on_stderr=kill_on_stderr,
+        _tor_config=config,
     )
-
-    # we set both to_delete and the shutdown events because this
-    # process might be shut down way before the reactor, but if the
-    # reactor bombs out without the subprocess getting closed cleanly,
-    # we'll want the system shutdown events triggered so the temporary
-    # files get cleaned up either way
-
-    # we don't want to delete the user's directories, just temporary
-    # ones this method created.
-    if not user_set_data_directory:
-        process_protocol.to_delete = [data_directory]
-        reactor.addSystemEventTrigger(
-            'before', 'shutdown',
-            functools.partial(delete_file_or_tree, data_directory)
-        )
-
-    try:
-        log.msg('Spawning tor process with DataDirectory', data_directory)
-        args = [tor_binary] + config_args
-        transport = reactor.spawnProcess(
-            process_protocol,
-            tor_binary,
-            args=args,
-            env={'HOME': data_directory},
-            path=data_directory if os.path.exists(data_directory) else None,
-        )
-        # FIXME? don't need rest of the args: uid, gid, usePTY, childFDs)
-        transport.closeStdin()
-
-    except RuntimeError as e:
-        return defer.fail(e)
-
-    if process_protocol.connected_cb:
-        return process_protocol.connected_cb
-    return defer.succeed(process_protocol)
+    defer.returnValue(tor.process)
 
 
 class TorConfigType(object):
@@ -690,7 +252,6 @@ class _ListWrapper(list):
         self.on_modify = on_modify_cb
 
     __setitem__ = _wrapture(list.__setitem__)
-    __setslice__ = _wrapture(list.__setslice__)
     append = _wrapture(list.append)
     extend = _wrapture(list.extend)
     insert = _wrapture(list.insert)
@@ -701,6 +262,10 @@ class _ListWrapper(list):
         return '_ListWrapper' + super(_ListWrapper, self).__repr__()
 
 
+if six.PY2:
+    setattr(_ListWrapper, '__setslice__', _wrapture(list.__setslice__))
+
+
 class HiddenServiceClientAuth(object):
     """
     Encapsulates a single client-authorization, as parsed from a
@@ -879,6 +444,15 @@ class EphemeralHiddenService(object):
     https://gitweb.torproject.org/torspec.git/tree/control-spec.txt#n1295
     '''
 
+    @classmethod
+    def _is_valid_keyblob(cls, key_blob_or_type):
+        try:
+            key_blob_or_type = nativeString(key_blob_or_type)
+        except (UnicodeError, TypeError):
+            return False
+        else:
+            return re.match(r'[^ :]+:[^ :]+$', key_blob_or_type)
+
     # XXX the "ports" stuff is still kind of an awkward API, especialy
     # making the actual list public (since it'll have
     # "80,127.0.0.1:80" instead of with a space
@@ -889,21 +463,22 @@ class EphemeralHiddenService(object):
     # XXX "auth" is unused (also, no Tor support I don't think?)
 
     def __init__(self, ports, key_blob_or_type='NEW:BEST', auth=[], ver=2):
-        if not isinstance(ports, types.ListType):
+        if not isinstance(ports, list):
             ports = [ports]
         # for "normal" HSes the port-config bit looks like "80
         # 127.0.0.1:1234" whereas this one wants a comma, so we leave
         # the public API the same and fix up the space. Or of course
         # you can just use the "real" comma-syntax if you wanted.
-        self._ports = map(lambda x: x.replace(' ', ','), ports)
-        self._key_blob = key_blob_or_type
+        self._ports = [x.replace(' ', ',') for x in ports]
+        if EphemeralHiddenService._is_valid_keyblob(key_blob_or_type):
+            self._key_blob = nativeString(key_blob_or_type)
+        else:
+            raise ValueError(
+                'key_blob_or_type must be a string in the formats '
+                '"NEW:<ALGORITHM>" or "<ALGORITHM>:<KEY>"')
         self.auth = auth  # FIXME ununsed
         # FIXME nicer than assert, plz
-        assert ' ' not in self._key_blob
-        assert isinstance(ports, types.ListType)
-        if not key_blob_or_type.startswith('NEW:') \
-           and (len(key_blob_or_type) > 825 or len(key_blob_or_type) < 820):
-            raise RuntimeError('Wrong size key-blob')
+        assert isinstance(ports, list)
 
     @defer.inlineCallbacks
     def add_to_tor(self, protocol):
@@ -917,8 +492,10 @@ class EphemeralHiddenService(object):
         ans = yield protocol.queue_command(cmd)
         ans = find_keywords(ans.split('\n'))
         self.hostname = ans['ServiceID'] + '.onion'
-        if self._key_blob == 'NEW:BEST':
+        if self._key_blob.startswith('NEW:'):
             self.private_key = ans['PrivateKey']
+        else:
+            self.private_key = self._key_blob
 
         log.msg('Created hidden-service at', self.hostname)
 
@@ -1071,6 +648,32 @@ def parse_client_keys(stream):
     return parser_state.keys
 
 
+def _endpoint_from_socksport_line(reactor, socks_config):
+    """
+    Internal helper.
+
+    Returns an IStreamClientEndpoint for the given config, which is of
+    the same format expected by the SOCKSPort option in Tor.
+    """
+    if socks_config.startswith('unix:'):
+        # XXX wait, can SOCKSPort lines with "unix:/path" still
+        # include options afterwards? What about if the path has a
+        # space in it?
+        return UNIXClientEndpoint(reactor, socks_config[5:])
+
+    # options like KeepAliveIsolateSOCKSAuth can be appended
+    # to a SocksPort line...
+    if ' ' in socks_config:
+        socks_config = socks_config.split()[0]
+    if ':' in socks_config:
+        host, port = socks_config.split(':', 1)
+        port = int(port)
+    else:
+        host = '127.0.0.1'
+        port = int(socks_config)
+    return TCP4ClientEndpoint(reactor, host, port)
+
+
 class TorConfig(object):
     """This class abstracts out Tor's config, and can be used both to
     create torrc files from nothing and track live configuration of a Tor
@@ -1110,19 +713,19 @@ class TorConfig(object):
     FIXME:
 
         - HiddenServiceOptions is special: GETCONF on it returns
-        several (well, two) values. Besides adding the two keys 'by
-        hand' do we need to do anything special? Can't we just depend
-        on users doing 'conf.hiddenservicedir = foo' AND
-        'conf.hiddenserviceport = bar' before a save() ?
+          several (well, two) values. Besides adding the two keys 'by
+          hand' do we need to do anything special? Can't we just depend
+          on users doing 'conf.hiddenservicedir = foo' AND
+          'conf.hiddenserviceport = bar' before a save() ?
 
         - once I determine a value is default, is there any way to
           actually get what this value is?
 
     """
 
-    @classmethod
+    @staticmethod
     @defer.inlineCallbacks
-    def from_protocol(cls, proto):
+    def from_protocol(proto):
         """
         This creates and returns a ready-to-go TorConfig instance from the
         given protocol, which should be an instance of
@@ -1143,7 +746,7 @@ class TorConfig(object):
         else:
             self._protocol = ITorControlProtocol(control)
 
-        self.unsaved = {}
+        self.unsaved = OrderedDict()
         '''Configuration that has been changed since last save().'''
 
         self.parsers = {}
@@ -1164,7 +767,7 @@ class TorConfig(object):
         if self.protocol:
             if self.protocol.post_bootstrap:
                 self.protocol.post_bootstrap.addCallback(
-                    self.bootstrap).addErrback(log.err)
+                    self.bootstrap).addErrback(self.post_bootstrap.errback)
             else:
                 self.bootstrap()
 
@@ -1173,6 +776,107 @@ class TorConfig(object):
 
         self.__dict__['_setup_'] = None
 
+    def socks_endpoint(self, reactor, port=None):
+        """
+        Returns a TorSocksEndpoint configured to use an already-configured
+        SOCKSPort from the Tor we're connected to. By default, this
+        will be the very first SOCKSPort.
+
+        :param port: a str, the first part of the SOCKSPort line (that
+            is, a port like "9151" or a Unix socket config like
+            "unix:/path". You may also specify a port as an int.
+
+        If you need to use a particular port that may or may not
+        already be configured, see the async method
+        :meth:`txtorcon.TorConfig.create_socks_endpoint`
+        """
+
+        if len(self.SocksPort) == 0:
+            raise RuntimeError(
+                "No SOCKS ports configured"
+            )
+
+        socks_config = None
+        if port is None:
+            socks_config = self.SocksPort[0]
+        else:
+            port = str(port)  # in case e.g. an int passed in
+            if ' ' in port:
+                raise ValueError(
+                    "Can't specify options; use create_socks_endpoint instead"
+                )
+
+            for idx, port_config in enumerate(self.SocksPort):
+                # "SOCKSPort" is a gnarly beast that can have a bunch
+                # of options appended, so we have to split off the
+                # first thing which *should* be the port (or can be a
+                # string like 'unix:')
+                if port_config.split()[0] == port:
+                    socks_config = port_config
+                    break
+        if socks_config is None:
+            raise RuntimeError(
+                "No SOCKSPort configured for port {}".format(port)
+            )
+
+        return _endpoint_from_socksport_line(reactor, socks_config)
+
+    @defer.inlineCallbacks
+    def create_socks_endpoint(self, reactor, socks_config):
+        """
+        Creates a new TorSocksEndpoint instance given a valid
+        configuration line for ``SocksPort``; if this configuration
+        isn't already in the underlying tor, we add it. Note that this
+        method may call :meth:`txtorcon.TorConfig.save()` on this instance.
+
+        Note that calling this with `socks_config=None` is equivalent
+        to calling `.socks_endpoint` (which is not async).
+
+        XXX socks_config should be .. i dunno, but there's fucking
+        options and craziness, e.g. default Tor Browser Bundle is:
+        ['9150 IPv6Traffic PreferIPv6 KeepAliveIsolateSOCKSAuth',
+        '9155']
+
+        XXX maybe we should say "socks_port" as the 3rd arg, insist
+        it's an int, and then allow/support all the other options
+        (e.g. via kwargs)
+
+        XXX we could avoid the "maybe call .save()" thing; worth it?
+        (actually, no we can't or the Tor won't have it config'd)
+        """
+
+        yield self.post_bootstrap
+
+        if socks_config is None:
+            if len(self.SocksPort) == 0:
+                raise RuntimeError(
+                    "socks_port is None and Tor has no SocksPorts configured"
+                )
+            socks_config = self.SocksPort[0]
+        else:
+            if not any([socks_config in port for port in self.SocksPort]):
+                # need to configure Tor
+                self.SocksPort.append(socks_config)
+                try:
+                    yield self.save()
+                except TorProtocolError as e:
+                    extra = ''
+                    if socks_config.startswith('unix:'):
+                        # XXX so why don't we check this for the
+                        # caller, earlier on?
+                        extra = '\nNote Tor has specific ownership/permissions ' +\
+                                'requirements for unix sockets and parent dir.'
+                    raise RuntimeError(
+                        "While configuring SOCKSPort to '{}', error from"
+                        " Tor: {}{}".format(
+                            socks_config, e, extra
+                        )
+                    )
+
+        defer.returnValue(
+            _endpoint_from_socksport_line(reactor, socks_config)
+        )
+
     # FIXME should re-name this to "tor_protocol" to be consistent
     # with other things? Or rename the other things?
     """
@@ -1183,6 +887,7 @@ class TorConfig(object):
     def _get_protocol(self):
         return self.__dict__['_protocol']
     protocol = property(_get_protocol)
+    tor_protocol = property(_get_protocol)
 
     def attach_protocol(self, proto):
         """
@@ -1202,13 +907,6 @@ class TorConfig(object):
             proto.post_bootstrap.addCallback(self.bootstrap)
         return self.__dict__['post_bootstrap']
 
-    def _update_proto(self, proto):
-        """
-        internal method, used by launch_tor to update the protocol after we're
-        set up.
-        """
-        self.__dict__['_protocol'] = proto
-
     def __setattr__(self, name, value):
         """
         we override this so that we can provide direct attribute
@@ -1441,7 +1139,9 @@ class TorConfig(object):
                 self._setup_hidden_services(servicelines)
                 continue
 
-            if value == 'Dependant':
+            # XXX for Virtual check that it's one of the *Ports things
+            # (because if not it should be an error)
+            if value in ('Dependant', 'Dependent', 'Virtual'):
                 continue
 
             # there's a thing called "Boolean+Auto" which is -1 for
@@ -1538,18 +1238,20 @@ class TorConfig(object):
 
     def config_args(self):
         '''
-        Returns an iterator of 2-tuples (config_name, value), one for
-        each configuration option in this config. This is more-or-less
-        and internal method, but see, e.g., launch_tor()'s
-        implementation if you thing you need to use this for
-        something.
+        Returns an iterator of 2-tuples (config_name, value), one for each
+        configuration option in this config. This is more-or-less an
+        internal method, but see, e.g., launch_tor()'s implementation
+        if you think you need to use this for something.
 
         See :meth:`txtorcon.TorConfig.create_torrc` which returns a
         string which is also a valid ``torrc`` file
-
         '''
 
-        for (k, v) in list(self.config.items()) + list(self.unsaved.items()):
+        everything = dict()
+        everything.update(self.config)
+        everything.update(self.unsaved)
+
+        for (k, v) in list(everything.items()):
             if type(v) is _ListWrapper:
                 if k.lower() == 'hiddenservices':
                     for x in v:
