@@ -4,11 +4,14 @@ from __future__ import absolute_import
 from __future__ import print_function
 from __future__ import with_statement
 
+from binascii import b2a_hex, hexlify
+
 from twisted.python import log
 from twisted.internet import defer
 from twisted.internet.interfaces import IProtocolFactory
 from twisted.internet.error import ConnectionDone
 from twisted.protocols.basic import LineOnlyReceiver
+from twisted.python.failure import Failure
 
 from zope.interface import implementer
 
@@ -17,6 +20,7 @@ from txtorcon.log import txtorlog
 
 from txtorcon.interface import ITorControlProtocol
 from .spaghetti import FSM, State, Transition
+from .util import maybe_coroutine
 
 import os
 import re
@@ -25,39 +29,6 @@ import base64
 DEFAULT_VALUE = 'DEFAULT'
 
 
-@defer.inlineCallbacks
-def connect(endpoint, password_function=lambda: None):
-    """
-    (experimental; details may change) This connects to the given
-    endpoint, presuming it is a Tor control connection and gives back
-    a TorControlProtocol instance. This does *not* reconnect if the
-    connection is dropped. This is the preferred entry point to
-    control a running Tor; if you need to *start* a Tor see
-    :ref:`launching_tor`.
-
-    See :meth:`txtorcon.TorState.from_protocol` to create a valid
-    :class:`txtorcon.TorState` instance from the connection.
-
-    :param endpoint: A Twisted IEndpoint, in practice either a
-        TCP4ClientEndpoint or a UnixClientEndpoint. By default, Tor
-        uses "tcp:localhost:9051" or "unix:/var/run/tor/control".
-
-    :param password_function: This is only consulted if the Tor to
-        which we connect does not have cookie authentication enabled. In
-        that case, this function is called to request a password. It may
-        return a Deferred.
-
-    :returns: A Deferred that fires with a ready-to-use
-        TorControlProtocol instance is returned, or Failure if a
-        connection can't be established.
-    """
-
-    factory = TorProtocolFactory(password_function=password_function)
-    proto = yield endpoint.connect(factory)
-    yield proto.post_bootstrap
-    defer.returnValue(proto)
-
-
 class TorProtocolError(RuntimeError):
     """
     Happens on 500-level responses in the protocol, almost certainly
@@ -139,9 +110,18 @@ class Event(object):
         self.callbacks.remove(cb)
 
     def got_update(self, data):
-        # print self.name,"got_update:",data
         for cb in self.callbacks:
-            cb(data)
+            try:
+                cb(data)
+            except Exception as e:
+                log.err(Failure())
+                log.err(
+                    "Notifying '{callback}' for '{name}' failed: {e}".format(
+                        callback=cb,
+                        name=self.name,
+                        e=e,
+                    )
+                )
 
 
 def unquote(word):
@@ -367,8 +347,7 @@ class TorControlProtocol(LineOnlyReceiver):
         GETINFO command. See :meth:`getinfo
         <txtorcon.TorControlProtocol.get_info>`
         """
-        info = ' '.join([str(x) for x in list(args)])
-        return self.queue_command('GETINFO %s' % info)
+        return self.queue_command('GETINFO %s' % ' '.join(args))
 
     def get_info_incremental(self, key, line_cb):
         """
@@ -482,16 +461,18 @@ class TorControlProtocol(LineOnlyReceiver):
         return self.queue_command('SIGNAL %s' % nm)
 
     def add_event_listener(self, evt, callback):
-        """:param evt: event name, see also
-        :var:`txtorcon.TorControlProtocol.events` .keys()
-
+        """
         Add a listener to an Event object. This may be called multiple
         times for the same event. If it's the first listener, a new
         SETEVENTS call will be initiated to Tor.
 
-        Currently the callback is any callable that takes a single
-        argument, that is the text collected for the event from the
-        tor control protocol.
+        :param evt: event name, see also
+            :attr:`txtorcon.TorControlProtocol.events` .keys(). These
+            event names are queried from Tor (with `GETINFO events/names`)
+
+        :param callback: any callable that takes a single argument
+             which receives the text collected for the event from the
+             tor control protocol.
 
         For more information on the events supported, see
         `control-spec section 4.1
@@ -505,9 +486,8 @@ class TorControlProtocol(LineOnlyReceiver):
         :Return: ``None``
 
         .. todo::
-            need an interface for the callback
-            show how to tie in Stem parsing if you want
-
+            - should have an interface for the callback
+            - show how to tie in Stem parsing if you want
         """
 
         if evt not in self.valid_events.values():
@@ -523,6 +503,13 @@ class TorControlProtocol(LineOnlyReceiver):
         return None
 
     def remove_event_listener(self, evt, cb):
+        """
+        The opposite of :meth:`TorControlProtocol.add_event_listener`
+
+        :param evt: the event name (or an Event object)
+
+        :param cb: the callback object to remove
+        """
         if evt not in self.valid_events.values():
             # this lets us pass a string or a real event-object
             try:
@@ -546,8 +533,18 @@ class TorControlProtocol(LineOnlyReceiver):
         return self.queue_command("PROTOCOLINFO 1")
 
     def authenticate(self, passphrase):
-        """Call the AUTHENTICATE command."""
-        return self.queue_command('AUTHENTICATE ' + passphrase.encode("hex"))
+        """
+        Call the AUTHENTICATE command.
+
+        Quoting torspec/control-spec.txt: "The authentication token
+        can be specified as either a quoted ASCII string, or as an
+        unquoted hexadecimal encoding of that same string (to avoid
+        escaping issues)."
+        """
+        if not isinstance(passphrase, bytes):
+            passphrase = passphrase.encode()
+        phrase = b2a_hex(passphrase)
+        return self.queue_command(b'AUTHENTICATE ' + phrase)
 
     def quit(self):
         """
@@ -569,6 +566,8 @@ class TorControlProtocol(LineOnlyReceiver):
         through this command.
         """
 
+        if not isinstance(cmd, bytes):
+            cmd = cmd.encode('ascii')
         d = defer.Deferred()
         self.commands.append((d, cmd, arg))
         self._maybe_issue_command()
@@ -583,9 +582,9 @@ class TorControlProtocol(LineOnlyReceiver):
         :api:`twisted.protocols.basic.LineOnlyReceiver` API
         """
 
-        self.debuglog.write(line + '\n')
+        self.debuglog.write(line + b'\n')
         self.debuglog.flush()
-        self.fsm.process(line)
+        self.fsm.process(line.decode('ascii'))
 
     def connectionMade(self):
         "Protocol API"
@@ -612,8 +611,9 @@ class TorControlProtocol(LineOnlyReceiver):
 
         firstline = rest[:rest.find('\n')]
         args = firstline.split()
-        if args[0] in self.events:
-            self.events[args[0]].got_update(rest[len(args[0]) + 1:])
+        name = args[0]
+        if name in self.events:
+            self.events[name].got_update(rest[len(name) + 1:])
             return
         # not considering this an error, as there's a slight window
         # after remove_event_listener is called (so the handler is
@@ -633,12 +633,12 @@ class TorControlProtocol(LineOnlyReceiver):
             (d, cmd, cmd_arg) = self.command
             self.defer = d
 
-            self.debuglog.write(cmd + '\n')
+            self.debuglog.write(cmd + b'\n')
             self.debuglog.flush()
 
-            data = cmd + '\r\n'
+            data = cmd + b'\r\n'
             txtorlog.msg("cmd: {}".format(data.strip()))
-            self.transport.write(data.encode('utf8'))
+            self.transport.write(data)
 
     def _auth_failed(self, fail):
         """
@@ -660,8 +660,8 @@ class TorControlProtocol(LineOnlyReceiver):
         server_nonce = base64.b16decode(kw['SERVERNONCE'])
         # FIXME put string in global. or something.
         expected_server_hash = hmac_sha256(
-            "Tor safe cookie authentication server-to-controller hash",
-            self._cookie_data + self.client_nonce + server_nonce
+            b"Tor safe cookie authentication server-to-controller hash",
+            self._cookie_data + self.client_nonce + server_nonce,
         )
 
         if not compare_via_hash(expected_server_hash, server_hash):
@@ -672,11 +672,11 @@ class TorControlProtocol(LineOnlyReceiver):
             )
 
         client_hash = hmac_sha256(
-            "Tor safe cookie authentication controller-to-server hash",
+            b"Tor safe cookie authentication controller-to-server hash",
             self._cookie_data + self.client_nonce + server_nonce
         )
         client_hash_hex = base64.b16encode(client_hash)
-        return self.queue_command('AUTHENTICATE %s' % client_hash_hex)
+        return self.queue_command(b'AUTHENTICATE ' + client_hash_hex)
 
     def _read_cookie(self, cookiefile):
         """
@@ -697,6 +697,7 @@ class TorControlProtocol(LineOnlyReceiver):
         what's supported.
         """
         methods = None
+        cookie_auth = False
         for line in protoinfo.split('\n'):
             if line[:5] == 'AUTH ':
                 kw = parse_keywords(line[5:].replace(' ', '\n'))
@@ -714,6 +715,7 @@ class TorControlProtocol(LineOnlyReceiver):
                 cookiefile = unescape_quoted_string(cookiefile)
                 try:
                     self._read_cookie(cookiefile)
+                    cookie_auth = True
                 except IOError as why:
                     txtorlog.msg("Reading COOKIEFILE failed: " + str(why))
                     if self.password_function and 'HASHEDPASSWORD' in methods:
@@ -732,29 +734,31 @@ class TorControlProtocol(LineOnlyReceiver):
                     "Got 'COOKIE' or 'SAFECOOKIE' method, but no 'COOKIEFILE'"
                 )
 
-        if 'SAFECOOKIE' in methods and self._cookie_data is not None:
-            txtorlog.msg("Using SAFECOOKIE authentication", cookiefile,
-                         len(self._cookie_data), "bytes")
-            self.client_nonce = os.urandom(32)
+        if cookie_auth:
+            if 'SAFECOOKIE' in methods:
+                txtorlog.msg("Using SAFECOOKIE authentication", cookiefile,
+                             len(self._cookie_data), "bytes")
+                self.client_nonce = os.urandom(32)
 
-            cmd = 'AUTHCHALLENGE SAFECOOKIE ' + \
-                  base64.b16encode(self.client_nonce)
-            d = self.queue_command(cmd)
-            d.addCallback(self._safecookie_authchallenge)
-            d.addCallback(self._bootstrap)
-            d.addErrback(self._auth_failed)
-            return
+                cmd = b'AUTHCHALLENGE SAFECOOKIE ' + \
+                      hexlify(self.client_nonce)
+                d = self.queue_command(cmd)
+                d.addCallback(self._safecookie_authchallenge)
+                d.addCallback(self._bootstrap)
+                d.addErrback(self._auth_failed)
+                return
 
-        elif 'COOKIE' in methods and self._cookie_data is not None:
-            txtorlog.msg("Using COOKIE authentication",
-                         cookiefile, len(self._cookie_data), "bytes")
-            d = self.authenticate(self._cookie_data)
-            d.addCallback(self._bootstrap)
-            d.addErrback(self._auth_failed)
-            return
+            elif 'COOKIE' in methods:
+                txtorlog.msg("Using COOKIE authentication",
+                             cookiefile, len(self._cookie_data), "bytes")
+                d = self.authenticate(self._cookie_data)
+                d.addCallback(self._bootstrap)
+                d.addErrback(self._auth_failed)
+                return
 
         if self.password_function and 'HASHEDPASSWORD' in methods:
             d = defer.maybeDeferred(self.password_function)
+            d.addCallback(maybe_coroutine)
             d.addCallback(self._do_password_authentication)
             d.addErrback(self._auth_failed)
             return
@@ -851,7 +855,6 @@ class TorControlProtocol(LineOnlyReceiver):
 
     def _is_continuation_line(self, line):
         "for FSM"
-        # print "isContinuationLine",self.code,line
         code = int(line[:3])
         if self.code and self.code != code:
             raise RuntimeError("Unexpected code %d, wanted %d" % (code,
@@ -860,7 +863,6 @@ class TorControlProtocol(LineOnlyReceiver):
 
     def _is_multi_line(self, line):
         "for FSM"
-        # print "isMultiLine",self.code,line,line[3] == '+'
         code = int(line[:3])
         if self.code and self.code != code:
             raise RuntimeError("Unexpected code %d, wanted %d" % (code,
@@ -898,7 +900,6 @@ class TorControlProtocol(LineOnlyReceiver):
 
     def _broadcast_response(self, line):
         "for FSM"
-        # print "BCAST",line
         if len(line) > 3:
             if self.code >= 200 and self.code < 300 and \
                self.command and self.command[2] is not None:
@@ -910,7 +911,9 @@ class TorControlProtocol(LineOnlyReceiver):
         else:
             resp = self.response
         self.response = ''
-        if self.code >= 200 and self.code < 300:
+        if self.code is None:
+            raise RuntimeError("No code set yet in broadcast response.")
+        elif self.code >= 200 and self.code < 300:
             if self.defer is None:
                 raise RuntimeError(
                     'Got a response, but didn\'t issue a command: "%s"' % resp
@@ -925,8 +928,6 @@ class TorControlProtocol(LineOnlyReceiver):
             self._handle_notify(self.code, resp)
             self.code = None
             return
-        elif self.code is None:
-            raise RuntimeError("No code set yet in broadcast response.")
         else:
             raise RuntimeError(
                 "Unknown code in broadcast response %d." % self.code
