@@ -11,13 +11,14 @@ import types
 import warnings
 
 from twisted.internet import defer
+from twisted.python.failure import Failure
 from twisted.internet.endpoints import TCP4ClientEndpoint
 from twisted.internet.endpoints import UNIXClientEndpoint
 from twisted.internet.interfaces import IReactorCore
 from twisted.internet.interfaces import IStreamClientEndpoint
 from zope.interface import implementer
 
-from txtorcon import TorProtocolFactory
+from txtorcon.torcontrolprotocol import TorProtocolFactory
 from txtorcon.stream import Stream
 from txtorcon.circuit import Circuit
 from txtorcon.router import Router, hashFromHexId
@@ -32,8 +33,9 @@ from txtorcon.interface import ICircuitListener
 from txtorcon.interface import ICircuitContainer
 from txtorcon.interface import IStreamListener
 from txtorcon.interface import IStreamAttacher
-from .spaghetti import FSM, State, Transition
-from .util import basestring
+from ._microdesc_parser import MicrodescriptorParser
+from .router import hexIdFromHash
+from .util import maybe_coroutine
 
 
 def _build_state(proto):
@@ -164,6 +166,24 @@ def flags_from_dict(kw):
     return flags
 
 
+def _extract_reason(kw):
+    """
+    Internal helper. Extracts a reason (possibly both reasons!) from
+    the kwargs for a circuit failed or closed event.
+    """
+    try:
+        # we "often" have a REASON
+        reason = kw['REASON']
+        try:
+            # ...and sometimes even have a REMOTE_REASON
+            reason = '{}, {}'.format(reason, kw['REMOTE_REASON'])
+        except KeyError:
+            pass  # should still be the 'REASON' error if we had it
+    except KeyError:
+        reason = "unknown"
+    return reason
+
+
 @implementer(ICircuitListener)
 @implementer(ICircuitContainer)
 @implementer(IRouterContainer)
@@ -214,7 +234,7 @@ class TorState(object):
         self.circuit_factory = Circuit
         self.stream_factory = Stream
 
-        self.attacher = None
+        self._attacher = None
         """If set, provides
         :class:`txtorcon.interface.IStreamAttacher` to attach new
         streams we hear about."""
@@ -258,127 +278,77 @@ class TorState(object):
         self.authorities = {}
 
         #: see set_attacher
-        self.cleanup = None
+        self._cleanup = None
 
-        class die(object):
-            __name__ = 'die'  # FIXME? just to ease spagetti.py:82's pain
-
-            def __init__(self, msg):
-                self.msg = msg
-
-            def __call__(self, *args):
-                raise RuntimeError(self.msg % tuple(args))
-
-        waiting_r = State("waiting_r")
-        waiting_w = State("waiting_w")
-        waiting_p = State("waiting_p")
-        waiting_s = State("waiting_s")
-
-        def ignorable_line(x):
-            x = x.strip()
-            return x in ['.', 'OK', ''] or x.startswith('ns/')
-
-        waiting_r.add_transition(Transition(waiting_r, ignorable_line, None))
-        waiting_r.add_transition(Transition(waiting_s, lambda x: x.startswith('r '), self._router_begin))
-        # FIXME use better method/func than die!!
-        waiting_r.add_transition(Transition(waiting_r, lambda x: not x.startswith('r '), die('Expected "r " while parsing routers not "%s"')))
-
-        waiting_s.add_transition(Transition(waiting_w, lambda x: x.startswith('s '), self._router_flags))
-        waiting_s.add_transition(Transition(waiting_s, lambda x: x.startswith('a '), self._router_address))
-        waiting_s.add_transition(Transition(waiting_r, ignorable_line, None))
-        waiting_s.add_transition(Transition(waiting_r, lambda x: not x.startswith('s ') and not x.startswith('a '), die('Expected "s " while parsing routers not "%s"')))
-        waiting_s.add_transition(Transition(waiting_r, lambda x: x.strip() == '.', None))
-
-        waiting_w.add_transition(Transition(waiting_p, lambda x: x.startswith('w '), self._router_bandwidth))
-        waiting_w.add_transition(Transition(waiting_r, ignorable_line, None))
-        waiting_w.add_transition(Transition(waiting_s, lambda x: x.startswith('r '), self._router_begin))  # "w" lines are optional
-        waiting_w.add_transition(Transition(waiting_r, lambda x: not x.startswith('w '), die('Expected "w " while parsing routers not "%s"')))
-        waiting_w.add_transition(Transition(waiting_r, lambda x: x.strip() == '.', None))
-
-        waiting_p.add_transition(Transition(waiting_r, lambda x: x.startswith('p '), self._router_policy))
-        waiting_p.add_transition(Transition(waiting_r, ignorable_line, None))
-        waiting_p.add_transition(Transition(waiting_s, lambda x: x.startswith('r '), self._router_begin))  # "p" lines are optional
-        waiting_p.add_transition(Transition(waiting_r, lambda x: x[:2] != 'p ', die('Expected "p " while parsing routers not "%s"')))
-        waiting_p.add_transition(Transition(waiting_r, lambda x: x.strip() == '.', None))
-
-        self._network_status_parser = FSM([waiting_r, waiting_s, waiting_w, waiting_p])
+        self._network_status_parser = MicrodescriptorParser(self._create_router)
 
         self.post_bootstrap = defer.Deferred()
         if bootstrap:
             self.protocol.post_bootstrap.addCallback(self._bootstrap)
             self.protocol.post_bootstrap.addErrback(self.post_bootstrap.errback)
 
-    def _router_begin(self, data):
-        args = data.split()
-        self._router = Router(self.protocol)
-        self._router.from_consensus = True
-        self._router.update(
-            args[1],                  # nickname
-            args[2],                  # idhash
-            args[3],                  # orhash
-            args[4] + ' ' + args[5],  # modified (like '%Y-%m-%f %H:%M:%S')
-            args[6],                  # ip address
-            args[7],                  # ORPort
-            args[8],                  # DirPort
+    def _create_router(self, **kw):
+        id_hex = hexIdFromHash(kw['idhash'])
+        try:
+            router = self.routers[id_hex]
+        except KeyError:
+            router = Router(self.protocol)
+            self.routers[router.id_hex] = router
+        router.from_consensus = True
+        router.update(
+            kw['nickname'],
+            kw['idhash'],
+            kw['orhash'],
+            kw['modified'],
+            kw['ip'],
+            kw['orport'],
+            kw['dirport'],
         )
+        router.flags = kw.get('flags', [])
+        if 'bandwidth' in kw:
+            router.bandwidth = kw['bandwidth']
+        if 'ip_v6' in kw:
+            router.ip_v6.extend(kw['ip_v6'])
 
-        if self._router.id_hex in self.routers:
+        if 'guard' in router.flags:
+            self.guards[router.id_hex] = router
+        if 'authority' in router.flags:
+            self.authorities[router.name] = router
+
+        if router.id_hex in self.routers:
             # FIXME should I do an update() on this one??
-            self._router = self.routers[self._router.id_hex]
-            return
+            router = self.routers[router.id_hex]
+        else:
+            if router.name in self.routers:
+                self.routers[router.name] = None
 
-        if self._router.name in self.routers_by_name:
-            self.routers_by_name[self._router.name].append(self._router)
+            else:
+                self.routers[router.name] = router
+
+        if router.name in self.routers_by_name:
+            self.routers_by_name[router.name].append(router)
 
         else:
-            self.routers_by_name[self._router.name] = [self._router]
+            self.routers_by_name[router.name] = [router]
 
-        if self._router.name in self.routers:
-            self.routers[self._router.name] = None
-
-        else:
-            self.routers[self._router.name] = self._router
-        self.routers[self._router.id_hex] = self._router
-        self.routers_by_hash[self._router.id_hex] = self._router
-        self.all_routers.add(self._router)
-
-    def _router_flags(self, data):
-        args = data.split()
-        self._router.flags = args[1:]
-        if 'guard' in self._router.flags:
-            self.guards[self._router.id_hex] = self._router
-        if 'authority' in self._router.flags:
-            self.authorities[self._router.name] = self._router
-
-    def _router_address(self, data):
-        """only for IPv6 addresses"""
-        self._router.ip_v6.append(data.split()[1].strip())
-
-    def _router_bandwidth(self, data):
-        args = data.split()
-        self._router.bandwidth = int(args[1].split('=')[1])
-
-    def _router_policy(self, data):
-        args = data.split()
-        self._router.policy = args[1:]
-        self._router = None
+        self.routers[router.id_hex] = router
+        self.routers_by_hash[router.id_hex] = router
+        self.all_routers.add(router)
 
     @defer.inlineCallbacks
     def _bootstrap(self, arg=None):
         "This takes an arg so we can use it as a callback (see __init__)."
 
         # update list of routers (must be before we do the
-        # circuit-status) note that we're feeding each line
-        # incrementally to a state-machine called
-        # _network_status_parser, set up in constructor. "ns" should
-        # be the empty string, but we call _update_network_status for
-        # the de-duplication of named routers
+        # circuit-status)
 
-        ns = yield self.protocol.get_info_incremental(
+        # look out! we're depending on get_info_incremental returning
+        # *lines*, which isn't documented -- but will be true because
+        # TorControlProtocol is a LineReceiver...
+        yield self.protocol.get_info_incremental(
             'ns/all',
-            self._network_status_parser.process
+            self._network_status_parser.feed_line,
         )
-        self._update_network_status(ns)
 
         # update list of existing circuits
         cs = yield self.protocol.get_info_raw('circuit-status')
@@ -453,28 +423,43 @@ class TorState(object):
     def set_attacher(self, attacher, myreactor):
         """
         Provide an :class:`txtorcon.interface.IStreamAttacher` to
-        associate streams to circuits. This won't get turned on until
-        after bootstrapping is completed. ('__LeaveStreamsUnattached'
-        needs to be set to '1' and the existing circuits list needs to
-        be populated).
+        associate streams to circuits.
+
+        You are Strongly Encouraged to **not** use this API directly,
+        and instead use :meth:`txtorcon.Circuit.stream_via` or
+        :meth:`txtorcon.Circuit.web_agent` instead. If you do need to
+        use this API, it's an error if you call either of the other
+        two methods.
+
+        This won't get turned on until after bootstrapping is
+        completed. ('__LeaveStreamsUnattached' needs to be set to '1'
+        and the existing circuits list needs to be populated).
         """
 
         react = IReactorCore(myreactor)
         if attacher:
-            self.attacher = IStreamAttacher(attacher)
+            if self._attacher is attacher:
+                return
+            if self._attacher is not None:
+                raise RuntimeError(
+                    "set_attacher called but we already have an attacher"
+                )
+            self._attacher = IStreamAttacher(attacher)
         else:
-            self.attacher = None
+            self._attacher = None
 
-        if self.attacher is None:
+        if self._attacher is None:
             d = self.undo_attacher()
-            if self.cleanup:
-                react.removeSystemEventTrigger(self.cleanup)
-                self.cleanup = None
+            if self._cleanup:
+                react.removeSystemEventTrigger(self._cleanup)
+                self._cleanup = None
 
         else:
             d = self.protocol.set_conf("__LeaveStreamsUnattached", "1")
-            self.cleanup = react.addSystemEventTrigger('before', 'shutdown',
-                                                       self.undo_attacher)
+            self._cleanup = react.addSystemEventTrigger(
+                'before', 'shutdown',
+                self.undo_attacher,
+            )
         return d
 
     # noqa
@@ -603,9 +588,10 @@ class TorState(object):
                     first = False
                 else:
                     cmd += ','
-                if isinstance(router, basestring) and len(router) == 40 \
+                # XXX should we really accept bytes here?
+                if isinstance(router, bytes) and len(router) == 40 \
                    and hashFromHexId(router):
-                    cmd += router
+                    cmd += router.decode('utf8')
                 else:
                     cmd += router.id_hex[1:]
         d = self.protocol.queue_command(cmd)
@@ -629,7 +615,7 @@ class TorState(object):
         (neither attaching it, nor telling Tor to attach it).
         """
 
-        if self.attacher is None:
+        if self._attacher is None:
             return None
 
         if stream.target_host is not None \
@@ -643,9 +629,10 @@ class TorState(object):
 
         # handle async or sync .attach() the same
         circ_d = defer.maybeDeferred(
-            self.attacher.attach_stream,
+            self._attacher.attach_stream,
             stream, self.circuits,
         )
+        circ_d.addCallback(maybe_coroutine)
 
         # actually do the attachment logic; .attach() can return 3 things:
         #    1. None: let Tor do whatever it wants
@@ -653,13 +640,9 @@ class TorState(object):
         #    3. Circuit instance: attach to the provided circuit
         def issue_stream_attach(circ):
             txtorlog.msg("circuit:", circ)
-            if circ is None:
+            if circ is None or circ is TorState.DO_NOT_ATTACH:
                 # tell Tor to do what it likes
-                return self.protocol.queue_command("ATTACHSTREAM %d 0" % stream.id)
-
-            elif circ is self.DO_NOT_ATTACH:
-                # do nothing; don't attach the stream
-                return
+                return self.protocol.queue_command(b"ATTACHSTREAM %d 0" % stream.id)
 
             else:
                 # should get a Circuit instance; check it for suitability
@@ -679,7 +662,7 @@ class TorState(object):
                     )
                 # we've got a valid Circuit instance; issue the command
                 return self.protocol.queue_command(
-                    "ATTACHSTREAM %d %d" % (stream.id, circ.id)
+                    b"ATTACHSTREAM %d %d" % (stream.id, circ.id)
                 )
 
         circ_d.addCallback(issue_stream_attach)
@@ -731,19 +714,26 @@ class TorState(object):
     def _update_network_status(self, data):
         """
         Used internally as a callback for updating Router information
-        from NS and NEWCONSENSUS events.
+        from NEWCONSENSUS events.
         """
 
-        self.all_routers = set()
-        for line in data.split('\n'):
-            self._network_status_parser.process(line)
+        # XXX why are we getting this with 0 data?
+        if len(data):
+            self.all_routers = set()
+            for line in data.split('\n'):
+                self._network_status_parser.feed_line(line)
+            self._network_status_parser.done()
 
         txtorlog.msg(len(self.routers_by_name), "named routers found.")
         # remove any names we added that turned out to have dups
+        remove_keys = set()
         for (k, v) in self.routers.items():
             if v is None:
                 txtorlog.msg(len(self.routers_by_name[k]), "dups:", k)
-                del self.routers[k]
+                remove_keys.add(k)
+
+        for k in remove_keys:
+            del self.routers[k]
 
         txtorlog.msg(len(self.guards), "GUARDs")
 
@@ -777,7 +767,6 @@ class TorState(object):
         from STREAM events.
         """
 
-        # print("stream_update", line)
         if line.strip() == 'stream-status=':
             # this happens if there are no active streams
             return
@@ -809,7 +798,6 @@ class TorState(object):
 
     event_map = {'STREAM': _stream_update,
                  'CIRC': _circuit_update,
-                 'NS': _update_network_status,
                  'NEWCONSENSUS': _update_network_status,
                  'ADDRMAP': _addr_map}
     """event_map used by add_events to map event_name -> unbound method"""
@@ -822,9 +810,14 @@ class TorState(object):
         for (event, func) in self.event_map.items():
             # the map contains unbound methods, so we bind them
             # to self so they call the right thing
+            try:
+                bound = types.MethodType(func, self, TorState)
+            except TypeError:
+                # python3
+                bound = types.MethodType(func, self)
             yield self.protocol.add_event_listener(
                 event,
-                types.MethodType(func, self, TorState)
+                bound,
             )
 
     # ICircuitContainer
@@ -928,16 +921,23 @@ class TorState(object):
     def circuit_destroy(self, circuit):
         "Used by circuit_closed and circuit_failed (below)"
         txtorlog.msg("circuit_destroy:", circuit.id)
-        for d in circuit._when_built:
-            d.errback(Exception("Destroying circuit; will never hit BUILT"))
+        circuit._when_built.fire(
+            Failure(Exception("Destroying circuit; will never hit BUILT"))
+        )
         del self.circuits[circuit.id]
 
     def circuit_closed(self, circuit, **kw):
         "ICircuitListener API"
         txtorlog.msg("circuit_closed", circuit)
+        circuit._when_built.fire(
+            Failure(Exception("Circuit closed ('{}')".format(_extract_reason(kw))))
+        )
         self.circuit_destroy(circuit)
 
     def circuit_failed(self, circuit, **kw):
         "ICircuitListener API"
         txtorlog.msg("circuit_failed", circuit, str(kw))
+        circuit._when_built.fire(
+            Failure(Exception("Circuit failed ('{}')".format(_extract_reason(kw))))
+        )
         self.circuit_destroy(circuit)
