@@ -14,13 +14,18 @@ import subprocess
 import ipaddress
 import struct
 import re
+import six
 
 from twisted.internet import defer
 from twisted.internet.interfaces import IProtocolFactory
-
 from twisted.internet.endpoints import serverFromString
+from twisted.web.http_headers import Headers
 
 from zope.interface import implementer
+from zope.interface import Interface
+
+if six.PY3:
+    import asyncio
 
 try:
     import GeoIP as _GeoIP
@@ -32,15 +37,35 @@ city = None
 country = None
 asn = None
 
-# XXX probably better to depend on and use "six" for py2/3 stuff?
-try:
-    unicode
-except NameError:
-    py3k = True
-    basestring = str
-else:
-    py3k = False
-    basestring = basestring
+
+def create_tbb_web_headers():
+    """
+    Returns a new `twisted.web.http_headers.Headers` instance
+    populated with tags to mimic Tor Browser. These include values for
+    `User-Agent`, `Accept`, `Accept-Language` and `Accept-Encoding`.
+    """
+    return Headers({
+        b"User-Agent": [b"Mozilla/5.0 (Windows NT 6.1; rv:45.0) Gecko/20100101 Firefox/45.0"],
+        b"Accept": [b"text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8"],
+        b"Accept-Language": [b"en-US,en;q=0.5"],
+        b"Accept-Encoding": [b"gzip, deflate"],
+    })
+
+
+def version_at_least(version_string, major, minor, micro, patch):
+    """
+    This returns True if the version_string represents a Tor version
+    of at least ``major``.``minor``.``micro``.``patch`` version,
+    ignoring any trailing specifiers.
+    """
+    parts = re.match(
+        r'^([0-9]+)\.([0-9]+)\.([0-9]+)\.([0-9]+).*$',
+        version_string,
+    )
+    for ver, gold in zip(parts.group(1, 2, 3, 4), (major, minor, micro, patch)):
+        if int(ver) < int(gold):
+            return False
+    return True
 
 
 def create_geoip(fname):
@@ -123,10 +148,12 @@ def maybe_ip_addr(addr):
     TODO consider explicitly checking for .exit or .onion at the end?
     """
 
+    if six.PY2 and isinstance(addr, str):
+        addr = unicode(addr)  # noqa
     try:
         return ipaddress.ip_address(addr)
     except ValueError:
-            pass
+        pass
     return str(addr)
 
 
@@ -190,7 +217,7 @@ def process_from_address(addr, port, torstate=None):
     proc = subprocess.Popen(['lsof', '-i', '4tcp@%s:%s' % (addr, port)],
                             stdout=subprocess.PIPE)
     (stdout, stderr) = proc.communicate()
-    lines = stdout.split('\n')
+    lines = stdout.split(b'\n')
     if len(lines) > 1:
         return int(lines[1].split()[1])
 
@@ -217,7 +244,7 @@ def compare_via_hash(x, y):
             hmac_sha256(CRYPTOVARIABLE_EQUALITY_COMPARISON_NONCE, y))
 
 
-class NetLocation:
+class NetLocation(object):
     """
     Represents the location of an IP address, either city or country
     level resolution depending on what GeoIP database was loaded. If
@@ -315,4 +342,155 @@ def unescape_quoted_string(string):
     # handeled as escape codes by string.decode('string-escape').
     # This is needed so e.g. '\x00' is not unescaped as '\0'
     string = re.sub(r'((?:^|[^\\])(?:\\\\)*)\\([^ntr0-7\\])', r'\1\2', string)
+    if six.PY3:
+        # XXX hmmm?
+        return bytes(string, 'ascii').decode('unicode-escape')
     return string.decode('string-escape')
+
+
+def default_control_port():
+    """
+    This returns a default control port, which respects an environment
+    variable `TX_CONTROL_PORT`. Without the environment variable, this
+    returns 9151 (the Tor Browser Bundle default).
+
+    You shouldn't use this in "normal" code, this is a convenience for
+    the examples.
+    """
+    try:
+        return int(os.environ['TX_CONTROL_PORT'])
+    except KeyError:
+        return 9151
+
+
+class IListener(Interface):
+    def add(callback):
+        """
+        Add a listener. The arguments to the callback are determined by whomever calls notify()
+        """
+
+    def remove(callback):
+        """
+        Add a listener. The arguments to the callback are determined by whomever calls notify()
+        """
+
+    def notify(*args, **kw):
+        """
+        Calls every listener with the given args and keyword-args.
+
+        XXX errors? just log?
+        """
+
+
+def maybe_coroutine(obj):
+    """
+    If 'obj' is a coroutine and we're using Python3, wrap it in
+    ensureDeferred. Otherwise return the original object.
+
+    (This is to insert in all callback chains from user code, in case
+    that user code is Python3 and used 'async def')
+    """
+    if six.PY3 and asyncio.iscoroutine(obj):
+        return defer.ensureDeferred(obj)
+    return obj
+
+
+@implementer(IListener)
+class _Listener(object):
+    """
+    Internal helper.
+    """
+
+    def __init__(self):
+        self._listeners = set()
+
+    def add(self, callback):
+        """
+        Add a callback to this listener
+        """
+        self._listeners.add(callback)
+
+    __call__ = add  #: alias for "add"
+
+    def remove(self, callback):
+        """
+        Remove a callback from this listener
+        """
+        self._listeners.remove(callback)
+
+    def notify(self, *args, **kw):
+        """
+        Calls all listeners with the specified args.
+
+        Returns a Deferred which callbacks when all the listeners
+        which return Deferreds have themselves completed.
+        """
+        calls = []
+
+        def failed(fail):
+            # XXX use logger
+            fail.printTraceback()
+
+        for cb in self._listeners:
+            d = defer.maybeDeferred(cb, *args, **kw)
+            d.addCallback(maybe_coroutine)
+            d.addErrback(failed)
+            calls.append(d)
+        return defer.DeferredList(calls)
+
+
+class _ListenerCollection(object):
+    """
+    Internal helper.
+
+    This collects all your valid event listeners together in one
+    object if you want.
+    """
+    def __init__(self, valid_events):
+        self._valid_events = valid_events
+        for e in valid_events:
+            setattr(self, e, _Listener())
+
+    def __call__(self, event, callback):
+        if event not in self._valid_events:
+            raise Exception("Invalid event '{}'".format(event))
+        getattr(self, event).add(callback)
+
+    def remove(self, event, callback):
+        if event not in self._valid_events:
+            raise Exception("Invalid event '{}'".format(event))
+        getattr(self, event).remove(callback)
+
+    def notify(self, event, *args, **kw):
+        if event not in self._valid_events:
+            raise Exception("Invalid event '{}'".format(event))
+        getattr(self, event).notify(*args, **kw)
+
+
+# similar to OneShotObserverList in Tahoe-LAFS
+class SingleObserver(object):
+    """
+    A helper for ".when_*()" sort of functions.
+    """
+    _NotFired = object()
+
+    def __init__(self):
+        self._observers = []
+        self._fired = self._NotFired
+
+    def when_fired(self):
+        d = defer.Deferred()
+        if self._fired is not self._NotFired:
+            d.callback(self._fired)
+        else:
+            self._observers.append(d)
+        return d
+
+    def fire(self, value):
+        if self._observers is None:
+            return  # raise RuntimeError("already fired") ?
+        self._fired = value
+        for d in self._observers:
+            d.callback(self._fired)
+        self._observers = None
+        return value  # so we're transparent if used as a callback
